from typing import *
import os
import sys

from phardwareitk.Extensions import C
from phardwareitk.Extensions.C import stdlib

do_print_exception = False

def set_exception_print(val:bool) -> None:
    """True for printing exception, False for otherwise"""
    global do_print_exception
    do_print_exception = val
    
def print_exception(exception) -> None:
    """Prints exception provided based on the do_print_exception, whose value is setted by set_exception_print function"""
    global do_print_exception
    if do_print_exception:
        print(exception)

SEEK_SET = 0
SEEK_CUR = 1
SEEK_END = 2

_IO_FILE = {
    "_IO_read_ptr": {
        "type": C.pointer,
        "ptr_type": C.char,
        "value": None
    },
    "_IO_read_end": {
        "type": C.pointer,
        "ptr_type": C.char,
        "value": None
    },
    "_IO_read_base": {
        "type": C.pointer,
        "ptr_type": C.char,
        "value": None
    },
    "_IO_write_base": {
        "type": C.pointer,
        "ptr_type": C.char,
        "value": None
    },
    "_IO_write_ptr": {
        "type": C.pointer,
        "ptr_type": C.char,
        "value": None
    },
    "_IO_write_end": {
        "type": C.pointer,
        "ptr_type": C.char,
        "value": None
    },
    "_IO_buf_base": {
        "type": C.pointer,
        "ptr_type": C.char,
        "value": None
    },
    "_IO_buf_end": {
        "type": C.pointer,
        "ptr_type": C.char,
        "value": None
    },
    "_fileno": {
        "type": C.Int,
        "value": None
    },
    "_blksize": {
        "type": C.Int,
        "value": None
    }
}
"""struct _IO_FILE {
   char  *_IO_read_ptr;   /* Current read pointer */
   char  *_IO_read_end;   /* End of get area. */
   char  *_IO_read_base;  /* Start of putback and get area. */
   char  *_IO_write_base; /* Start of put area. */
   char  *_IO_write_ptr;  /* Current put pointer. */
   char  *_IO_write_end;  /* End of put area. */
   char  *_IO_buf_base;   /* Start of reserve area. */
   char  *_IO_buf_end;   /* End of reserve area. */
   int   _fileno;
   int   _blksize;
};

typedef struct _IO_FILE FILE;"""

FILE = C.struct[_IO_FILE]

def _get_string(ptr: Union[C.pointer[C.char], str, None]) -> str:
    if ptr is None: return None
    if isinstance(ptr, str): return ptr
    return ptr.deref()

def fopen(path:Union[C.pointer[C.char], str], mode:Union[C.pointer[C.char], str]) -> Union[C.pointer[FILE], C.Int]:
    """fopen - Open a file and return a FILE struct pointer

    Arguments:
        path (char*):
            Path to the file
        mode (char*):
            Mode to open the file with (e.g. 'r', 'w', 'rb', ...)

    Returns: Pointer to FILE (_IO_FILE) struct or -1 on failure, -2 on file not found, -3 on permission error"""
    C.push_frame()
    filename = _get_string(path)
    mode = _get_string(mode)

    if "b" not in mode:
        mode += "b"

    fd = None

    try:
        fd = open(filename, mode)
    except PermissionError as pe:
        print_exception(pe)
        return C.Int(-3)

    if fd is None:
        return C.Int(-1)

    buffer_size = 4096 # Default
    statvfs = None
    block_size = None

    try:
        statvfs = os.statvfs(filename)
        block_size = statvfs.f_bsize
    except:
        block_size = buffer_size

    # allocate memory for buffer
    buf_base = stdlib.malloc(C.size_t(buffer_size))

    # Allocate FILE struct
    file = C.struct(_IO_FILE)
    # set read values first
    file._IO_read_ptr = buf_base
    file._IO_read_end = buf_base # empty buffer at the start
    file._IO_read_base = buf_base

    # set write values
    file._IO_write_base = buf_base
    file._IO_write_ptr = buf_base
    file._IO_write_end = buf_base.ptr_addr + buffer_size

    # set buffer values
    file._IO_buf_base = buf_base
    file._IO_buf_end = buf_base.ptr_addr + buffer_size

    # set file descriptor and block size
    try:
        file._fileno = C.Int(fd.fileno())
    except Exception as e:
        print_exception(e)
        file._fileno = C.Int(-1)

    file._blksize = C.Int(block_size)

    # Store the file descriptor in a global dictionary to keep track of open files
    if "_open_files" not in globals():
        global _open_files
        _open_files = {}
    _open_files[fd.fileno()] = fd

    C.pop_frame()
    return C.pointer(file, C.struct) # Return

def fclose(file_:C.pointer[FILE]) -> C.Int:
    """fclose - Close a file and free its memory

    Arguments:
        file (FILE*):
            Pointer to the FILE struct

    Returns: 0 on success, -1 on failure"""
    try:
        C.push_frame()
        # Dereference the pointer
        file: C.struct = C.struct.from_address(_IO_FILE, file_.ptr_addr)
        fileno = file._fileno.value

        # Get the file descriptor from the global dictionary
        fd:TextIO = _open_files.get(fileno, None)

        if fd is None:
            return C.Int(-1)

        # Close the file
        fd.close()

        # Free the buffer memory
        stdlib.free(file_) # Free the FILE struct memory
        del file
        del file_

        # Remove the file descriptor from the global dictionary
        del _open_files[fileno]
        C.pop_frame()
        return C.Int(0)
    except Exception as e:
        print_exception(e)
        return C.Int(-1)

def ftell(file: C.pointer[FILE]) -> C.size_t:
    """ftell - Get the current position in the file

    Arguments:
        file (FILE*):
            Pointer to the FILE struct

    Returns: Current position in the file or -1 on failure"""
    try:
        # Dereference the pointer
        C.push_frame()
        file_struct: C.struct = C.struct.from_address(_IO_FILE, file.ptr_addr)
        fd = _open_files.get(file_struct._fileno.value)
        if not fd:
            del file_struct
            return C.Int(-1)
        out = fd.tell()
        del file_struct
        C.pop_frame()
        return C.size_t(out)
    except Exception as e:
        print_exception(e)
        return C.Int(-1)

def fflush(file: C.pointer[FILE]) -> C.Int:
    """fflush - Flush the output buffer of a file

    Arguments:
        file (FILE*):
            Pointer to the FILE struct

    Returns: 0 on success, -1 on failure"""
    try:
        # Dereference the pointer
        file_struct: C.struct = C.struct.from_address(_IO_FILE, file.ptr_addr)
        fd:TextIO = _open_files.get(file_struct._fileno.value, None)

        if fd is None:
            return C.Int(-1)

        # Flush the file
        fd.flush()
        del file_struct
        C.pop_frame()
        return C.Int(0)
    except Exception as e:
        print_exception(e)
        return C.Int(-1)

def fseek(file: C.pointer[FILE], offset: C.Int, whence: C.Int) -> C.Int:
    """fseek - Set the file position indicator for the stream

    Arguments:
        file (FILE*):
            Pointer to the FILE struct
        offset (int):
            Offset to set the position to
        whence (int):
            Position from which to set the offset (e.g. SEEK_SET, SEEK_CUR, SEEK_END)

    Returns: 0 on success, -1 on failure"""
    try:
        # Dereference the pointer
        C.push_frame()
        file_struct: C.struct = C.struct.from_address(_IO_FILE, file.ptr_addr)
        fd:TextIO = _open_files.get(file_struct._fileno.value, None)

        if fd is None:
            return Int(-1)

        fd.seek(offset.value, whence.value)
        del file_struct
        C.pop_frame()
        return C.Int(0)
    except Exception as e:
        print_exception(e)
        return C.Int(-1)

def fread(dest: C.pointer[C.void], size: C.size_t, nmemb: C.size_t, file_ptr: C.pointer[FILE]) -> C.Int:
    """fread - Read data from a file

    Args:
        dest (void*): The destination pointer where data will be stored
        size (size_t): size of each element to read
        nmemb (size_t): number of elements to read
        file_ptr (FILE*): Pointer to the FILE struct

    Returns:
        int: Number of elements successfully read, or -1 on failure
    """
    size = size.value
    nmemb = nmemb.value

    try:
        C.push_frame()
        # Dereference the pointer
        file: C.struct = C.struct.from_address(_IO_FILE, file_ptr.ptr_addr)
        total_bytes = size * nmemb

        fileno = file._fileno.value
        fd:TextIO = _open_files.get(fileno, None)
        if fd is None:
            return C.Int(-1)

        data = fd.read(total_bytes)
        if not data:
            return C.Int(0)

        if isinstance(data, str): data = data.encode("utf-8")

        # Write data to the destination pointer
        C.write_mem(data, size.value, dest.ptr_addr)

        # Update the read pointer
        buf_base = file._IO_buf_base
        file._IO_read_ptr = C.pointer(buf_base.ptr_addr + len(data), C.char)
        file._IO_read_end = C.pointer(buf_base.ptr_addr + total_bytes, C.char)

        del file

        C.pop_frame()
        return C.Int(len(data) // size)

    except Exception as e:
        print_exception(e)
        return C.Int(-1)

def fwrite(src: Union[C.pointer[C.void], str, bytes], size: C.size_t, nmemb: C.size_t, file_ptr: C.pointer[FILE], chunk_size:Optional[C.Int]=None) -> C.Int:
    """fwrite - Write data to a file
    
    Args:
        src (void*): The source pointer containing data to write
        size (size_t): size of each element to write
        nmemb (size_t): number of elements to write
        file_ptr (FILE*): Pointer to the FILE struct
        chunk_size (int): Writing files in chunk to prevent overuse of memory. Defaults to 4096 bytes / 4KB

    Returns:
        int: Number of elements successfully written, or -1 on failure
    """
    size = size.value
    nmemb = nmemb.value
    chunk_size = chunk_size.value if chunk_size else 4096

    try:
        C.push_frame()
        # Dereference the pointer
        file:C.struct = C.struct.from_address(_IO_FILE, file_ptr.ptr_addr)
        total_bytes = size * nmemb

        fileno = file._fileno.value
        fd:TextIO = _open_files.get(fileno, None)
        if fd is None:
            return C.Int(-1)
        
        written = 0
        srcaddr = src.ptr_addr
        
        offset = 0
        print("memaddr: ", srcaddr)
        for i in range(nmemb):
            buf = C.read_mem(size, srcaddr + offset)

            fd.write(buf)
            written += 1
            offset += size

        fd.flush()
        os.fsync(fd.fileno())
      
        # Update the write pointer
        buf_base = file._IO_buf_base
        file._IO_write_ptr = C.pointer(buf_base.ptr_addr + written, C.char)

        del file
        
        C.pop_frame()
        return C.Int(written)

    except Exception as e:
        print_exception(e)
        return C.Int(-1)

def remove(filename: Union[C.pointer[C.char], str]) -> C.Int:
    """remove - Remove a file

    Arguments:
        filename (char*):
            Path to the file to remove

    Returns: 0 on success, -1 on failure"""
    try:
        C.push_frame()
        os.remove(_get_string(filename))
        C.pop_frame()
        return C.Int(0)
    except FileNotFoundError as fe:
        print_exception(fe)
        return C.Int(-2)
    except PermissionError as pe:
        print_exception(pe)
        return C.Int(-3)
    except Exception as e:
        print_exception(e)
        return C.Int(-1)

def rename(old_filename: Union[C.pointer[C.char], str], new_filename: Union[C.pointer[C.char], str]) -> C.Int:
    """rename - Rename a file

    Arguments:
        old_filename (char*):
            Path to the file to rename
        new_filename (char*):
            New path for the file

    Returns: 0 on success, -1 on failure"""
    try:
        C.push_frame()
        os.rename(_get_string(old_filename), _get_string(new_filename))
        C.pop_frame()
        return C.Int(0)
    except FileNotFoundError as fe:
        print_exception(fe)
        return C.Int(-2)
    except PermissionError as pe:
        print_exception(pe)
        return C.Int(-3)
    except Exception as e:
        print_exception(e)
        return C.Int(-1)




import os
import sys
import time
import struct

# NOTE: This is PBFS 64-bit version, for 128-bit version that supports ~356,000,000 Yottabytes please download the C/ASM version of pbfs! thanks a lot.
MPATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..")

if not MPATH in sys.path:
    sys.path.append(MPATH)

from Extensions import C
from Extensions.C import stdlib
from Extensions.C import stdio
from Extensions.C import stdint

PBFS_HEADER = {
    "Magic": {"type": C.array, "array_type": C.char, "array_len": 6, "value": None},
    "Block_Size": {"type": stdint.uint32_t, "value": None},
    "Total_Blocks": {"type": stdint.uint32_t, "value": None},
    "Disk_Name": {"type": C.array, "array_type": C.char, "array_len": 24, "value": None},
    "TimeStamp": {"type": stdint.uint64_t, "value": None},
    "Version": {"type": stdint.uint32_t, "value": None},
    "First_Boot_Timestamp": {"type": stdint.uint64_t, "value": None},
    "OS_BootMode": {"type": stdint.uint16_t, "value": None},
    "Entries": {"type": stdint.uint32_t, "value": None},
}

"""typedef struct {
	char Magic[6]; // PBFS\x00\x00
	uint32_t Block_Size; // 512 bytes by default (BIOS BLOCK SIZE)
	uint32_t Total_Blocks;  // 2048 blocks for 1MB
	char Disk_Name[24]; // Just the name of the disk for the os
	uint64_t Timestamp; // Timestamp
	uint32_t Version; // Version
	uint64_t First_Boot_Timestamp; // First boot timestamp
	uint16_t OS_BootMode; // Again optional but there for furture use!
	uint32_t FileTableOffset; // Offset of the file table (first data block)
	uint32_t Entries; // Number of entries in the file table
} __attribute__((packed)) PBFS_Header; // Total = 68 bytes"""

PBFS_FILE_TABLE_ENTRY = {
    "Name": {"type": C.array, "array_type": C.char, "array_len": 150, "value": None},
    "File_Data_Offset": {"type": stdint.uint64_t, "value": None},
    "Permission_Table_Offset": {"type": stdint.uint64_t, "value": None},
    "Block_Span": {"type": stdint.uint64_t, "value": None},
}

"""typedef struct {
	char Name[150]; // Name of the file
	uint64_t File_Data_Offset; // File data offset
	uint64_t Permission_Table_Offset; // Permission table offset
	uint64_t Block_Span; // File Block Span
} __attribute__((packed)) PBFS_FileTableEntry; // Total = 24 bytes"""

PBFS_PERMISSION_TABLE_ENTRY = {
    "Read": {"type": stdint.uint16_t, "value": None},
    "Write": {"type": stdint.uint16_t, "value": None},
    "Executable": {"type": stdint.uint16_t, "value": None},
    "Listable": {"type": stdint.uint16_t, "value": None},
    "Hidden": {"type": stdint.uint16_t, "value": None},
    "Full_Control": {"type": stdint.uint16_t, "value": None},
    "Delete": {"type": stdint.uint16_t, "value": None},
    "Special_Access": {"type": stdint.uint16_t, "value": None},
    "File_Tree_Offset": {"type": stdint.uint32_t, "value": None},
}

"""typedef struct {
	uint16_t Read; // Read Permission
	uint16_t Write; // Write Permission
	uint16_t Executable; // Executable Permission
	uint16_t Listable; // Listable Permission
	uint16_t Hidden; // Hidden Permission
	uint16_t Full_Control; // System File Permission
	uint16_t Delete; // Delete Permission
	uint16_t Special_Access; // Special Access
	uint32_t File_Tree_Offset; // Offset of the file tree
} __attribute__((packed)) PBFS_PermissionTableEntry; // Total = 20 bytes"""

PBFS_FILE_TREE_ENTRY = {
    "Name": {"type": C.array, "array_type": C.char, "array_len": 20, "value": None}
}

"""typedef struct {
	char Name[20]; // Name of the file (unique id is also accepted, it is 20 bytes because the file paths are not used here but rather the unique id is used).
} __attribute__((packed)) PBFS_FileTreeEntry; // Total = 20 bytes"""

PBFS_DAP = {
    "size": {"type": stdint.uint8_t, "value": None},
    "reserved": {"type": stdint.uint8_t, "value": None},
    "sector_count": {"type": stdint.uint16_t, "value": None},
    "offset": {"type": stdint.uint16_t, "value": None},
    "segment": {"type": stdint.uint16_t, "value": None},
    "lba": {"type": stdint.uint64_t, "value": None},
}

"""typedef struct {
	uint8_t size;
	uint8_t reserved;
	uint16_t sector_count;
	uint16_t offset;
	uint16_t segment;
	uint64_t lba;
} __attribute__((packed)) PBFS_DAP;"""

PBFS_PERMISSIONS = {
    "Read": {"type": stdint.uint16_t, "value": None},
    "Write": {"type": stdint.uint16_t, "value": None},
    "Executable": {"type": stdint.uint16_t, "value": None},
    "Listable": {"type": stdint.uint16_t, "value": None},
    "Hidden": {"type": stdint.uint16_t, "value": None},
    "Full_Control": {"type": stdint.uint16_t, "value": None},
    "Delete": {"type": stdint.uint16_t, "value": None},
    "Special_Access": {"type": stdint.uint16_t, "value": None},
}

"""typedef struct {
	uint16_t Read; // Read Permission
	uint16_t Write; // Write Permission
	uint16_t Executable; // Executable Permission
	uint16_t Listable; // Listable Permission
	uint16_t Hidden; // Hidden Permission
	uint16_t Full_Control; // System File Permission
	uint16_t Delete; // Delete Permission
	uint16_t Special_Access; // Special Access
} __attribute__((packed)) PBFS_Permissions; // Total = 16 bytes"""

PBFS_LAYOUT = {
    "Header_Start": {"type": stdint.uint64_t, "value": None},
    "Header_End": {"type": stdint.uint64_t, "value": None},
    "Header_BlockSpan": {"type": stdint.uint64_t, "value": None},
    "Bitmap_Start": {"type": stdint.uint64_t, "value": None},
    "Bitmap_BlockSpan": {"type": stdint.uint64_t, "value": None},
    "Bitmap_End": {"type": stdint.uint64_t, "value": None},
    "Data_Start": {"type": stdint.uint64_t, "value": None},
}

"""typedef struct {
	uint64_t Header_Start;
	uint64_t Header_End;
	uint64_t Header_BlockSpan;
	uint64_t Bitmap_Start;
	uint64_t Bitmap_BlockSpan;
	uint64_t Bitmap_End;
	uint64_t Data_Start;
} __attribute__((packed)) PBFS_Layout;"""

DRIVE_PARAMETERS = {
    "size": {"type": stdint.uint16_t, "value": None},
    "flags": {"type": stdint.uint16_t, "value": None},
    "cylinders": {"type": stdint.uint32_t, "value": None},
    "heads": {"type": stdint.uint32_t, "value": None},
    "sectors_per_track": {"type": stdint.uint32_t, "value": None},
    "total_sectors": {"type": stdint.uint64_t, "value": None},
    "bytes_per_sector": {"type": stdint.uint16_t, "value": None},
    "reserved": {"type": C.array, "array_type": stdint.uint8_t, "array_len": 6, "value": None},
}

"""typedef struct {
	uint16_t size;              // 0x00 - Must be 0x1E (30)
	uint16_t flags;             // 0x02
	uint32_t cylinders;         // 0x04 - reserved or zero
	uint32_t heads;             // 0x08 - reserved or zero
	uint32_t sectors_per_track; // 0x0C - reserved or zero
	uint64_t total_sectors;     // 0x10 - important!
	uint16_t bytes_per_sector;  // 0x18 - important!
	uint8_t  reserved[6];       // 0x1A - reserved
} __attribute__((packed)) DriveParameters;"""

PBFS_FILE_LIST_ENTRY = {
    "Name": {"type": C.pointer, "ptr_type": C.char, "value": None},
    "lba": {"type": stdint.uint64_t, "value": None},
}

"""typedef struct {
	char* name;
	uint64_t lba;
} PBFS_FileListEntry;"""

memsize = 64

def round_up_to_block(size: int, block_size: int = 512) -> int:
    return ((size + block_size - 1) // block_size) * block_size

def validate_disk(path: str, block_size: int = 512) -> bool:
    """Validates the Drive"""
    C.push_frame()
    if not os.path.exists(path):
        return False

    drive: C.pointer[stdio.FILE] = stdio.fopen(path, "rb")
    stdio.fseek(drive, C.Int(0), C.Int(stdio.SEEK_END))
    size = stdio.ftell(drive)
    stdio.fseek(drive, C.Int(0), C.Int(stdio.SEEK_SET))
    buffer_ = stdlib.malloc(C.size_t(block_size))
    buffer_.cast(C.Char)
    stdio.fseek(drive, C.Int(block_size), C.Int(stdio.SEEK_SET))
    stdio.fread(buffer_, C.size_t(block_size), C.size_t(1), drive)
    err = C.struct.from_address(PBFS_HEADER, buffer_.ptr_addr)
    if err == -1:
        print("Incorrect data in file: Validation Failed!")
        stdlib.free(buffer_)
        stdio.fclose(drive)
        C.pop_frame()
        return False

    magic = err.Magic
    magic_val = C.pointer(magic.address, C.char).deref()
    if not magic_val == b"PBFS\x00\x00":
        print("Signature doesn't Match! Validation failed!")
        stdlib.free(buffer_)
        stdio.fclose(drive)
        C.pop_frame()
        return False

    stdlib.free(buffer_)
    stdio.fclose(drive)
    C.pop_frame()
    return True

def format_disk(
    path: str,
    total_blocks: int = 2048,
    block_size: int = 512,
    disk_name: bytes = b"SSD-PBFS-VIRTUAL",
) -> int:
    """Formates the Drive"""
    C.reset_mem(memsize) # Always ensure clean slate
    C.push_frame()
    print("Formatting disk...")
    file: C.pointer[stdio.FILE] = stdio.fopen(path, "wb+")

    if isinstance(file, C.Int):
        print("Opening the file failed:", file)
        return -1

    try:
        lba_buff = stdlib.malloc(C.size_t(block_size))
        if not lba_buff: 
            print("Alloc failed!")
            return -1
        # Now we format it
        print("Formatting...")
        C.write_mem(b"\x00" * block_size, block_size, lba_buff.ptr_addr)
        stdio.fwrite(lba_buff, C.size_t(block_size), C.size_t(1), file)

        print("Writing PBFS Header...")
        PBFS_Header = C.struct.from_address(PBFS_HEADER, lba_buff.ptr_addr)
        PBFS_Header.Magic = b"PBFS\x00\x00"
        PBFS_Header.Block_Size = stdint.uint32_t(block_size)
        PBFS_Header.Total_Blocks = stdint.uint32_t(total_blocks)
        PBFS_Header.Disk_Name = disk_name
        PBFS_Header.TimeStamp = stdint.uint64_t(int(time.time()))
        PBFS_Header.Version = stdint.uint32_t(1)
        PBFS_Header.Entries = stdint.uint32_t(0)
        PBFS_Header.First_Boot_Timestamp = stdint.uint64_t(0)
        PBFS_Header.OS_BootMode = stdint.uint16_t(1)

        print("lba addr: ", lba_buff.ptr_addr)

        print("Writing...")
        stdio.fwrite(lba_buff, C.size_t(block_size), C.size_t(1), file)
        C.write_mem(b"\x00" * block_size, block_size, lba_buff.ptr_addr)
        
        # Create bitmap
        print("Creating Bitmap")
        
        def make_bitmap(total_blocks_, reserved_blocks=3):
            bytes_needed = (total_blocks_ + 7) // 8
            bitmap = bytearray(bytes_needed)
            
            for block in range(reserved_blocks):
                byte_index = block // 8
                bit_index = block % 8
                bitmap[byte_index] |= (1 << bit_index)
                
            return bitmap
        
        bmap = make_bitmap(total_blocks, 3)
        space_needed = (total_blocks + 7) // 8
        
        C.write_mem(bmap, space_needed, lba_buff.ptr_addr)
        stdio.fwrite(lba_buff, C.size_t(block_size), C.size_t(space_needed), file)
        
        stdio.fwrite(b"\x00" * block_size, C.size_t(block_size), C.size_t(total_blocks - (1 + space_needed)), file)
        
        print("Done formatting disk...")
        stdio.fflush(file)
    except Exception as e:
        print(f"Exception: {e}")
        stdio.fclose(file)
        stdlib.free(lba_buff)
        print("FAILED!")
        C.pop_frame()
        return -1
    stdio.fclose(file)
    stdlib.free(lba_buff)

    print("Finished!")
    C.pop_frame()

    return 0


def increase_memsize(size_: int):
    """Increase the default memory size for this file"""
    global memsize
    memsize = size_
    C.reset_mem(size_)

def initialize_pbfs(size_: int):
    global memsize
    memsize = size_
    C.initialize(size_)


class PBFS:
    """Python Block File System / Pheonix Block File System"""

    def __init__(self, drive: str, block_size: int=512, total_blocks: int=2048) -> None:
        pass


from typing import *

def cmd_aos_help(print_:bool, flush:bool=False, endl:str="\n") -> Optional[None, str]:
	"""AOS cmd help func"""
	help_ = """Usage: aos <command> <params>
	
aos is a command by AOS which allows users to run commands that interact with the system directtly which is not possible normally.
	
Basic Commands:
	wipe-drive: aos wipe-drive
		Wipes and then partions drive in PBFS (Python Based File System/Pheonix Base File System)
		If you use this command all data will be lost excluding the OS itself.
	
	cd-usr: aos cd-usr <user>
		Changes the default user (admin) to another valid user
		
	mk-usr: aos mk-usr <username> <options>
		Makes a new user with the specified username.
		
		OPTIONS:
			-p: defines a password for the user created
	
	rm-usr: aos rm-usr <user>
		Removes the specified user from the system
		
	rn-usr: aos rn-usr <user> <username>
		Renames a user with the specified username
		
Display Commands:
	display-mode: aos display-mode <mode as a number>
		Changes the default display mode.
		
		Modes:
			0 -> Command Line Interface
			1 -> VGA
			2 -> FrameBuffer
			3 -> PHardwareITK.GUI (Not recommended as it is a sdl2 package and may be slower to run)
			
		Default Modes:
			AOS:Penv:Potato -> Command Line Interface
			AOS:Penv:Full -> VGA (at booting), Framebuffer (For the rest of the time)
"""

	if not print_:
		return help_
		
	print(help_, flush, end=endl)



import os
import platform

from typing import * # type: ignore

class PIenum():
    def __init__(self, value:int=None) -> None:
        if not value:
            self.value = 0
        else:
            self.value = value
        self.entries = []
        self.start_value = 0
        self.enum_dict = {}
        self.enum = []
        self.reverse_enum_dict = {}

    def create_enum(self, entries:list[str], start_value:int=0) -> None:
        self.entries = entries
        self.value = (len(self.entries) - 1) + start_value
        self.start_value = start_value
        self.enum_dict = {name: idx + start_value for idx, name in enumerate(entries)}
        self.reverse_enum_dict = {idx + start_value: name for idx, name in enumerate(entries)}

        self.enum = list(self.enum_dict.values())

    def access(self, name:str) -> int:
        return self.enum_dict.get(name, None)
    
    def reverse_access(self, val:int) -> str:
        return self.reverse_enum_dict.get(val, None)
    
    def __getitem__(self, item:Union[str, int]) -> Any:
        if isinstance(item, str):
            return self.enum_dict.get(item, None)
        elif isinstance(item, int):
            try:
                return self.enum[item]
            except IndexError:
                return None
        return None

    def __repr__(self) -> str:
        return f"PIenum({self.enum_dict})"

class PIint(int):
    def __init__(self, value:int=0) -> None:
        self.value = value
        super().__init__(value)

    def __repr__(self) -> str:
        return f"PIint({self.value})"

class PIuint(int):
    def __init__(self, value:int=0) -> None:
        if value < 0:
            raise ValueError("PIuint cannot be negative")
        self.value = value
        super().__init__(value)

    def __repr__(self) -> str:
        return f"PIuint({self.value})"

class PIstring(str):
    def __init__(self, value:str="") -> None:
        self.value = value
        super().__init__(value)

    def __repr__(self) -> str:
        return f'PIstring("{self.value}")'
    
_pionevent_types_str = [
    # Window lifecycle
    "CREATE",
    "DESTROY",
    "SHOW",
    "HIDE",
    "FOCUS_GAINED",
    "FOCUS_LOST",
    "RESIZE",
    "MOVE",
    "CLOSE",
    "QUIT",
    "MINIMIZE",
    "MAXIMIZE",
    "RESTORE",
    "ENTER",
    "LEAVE",

    # Mouse
    "MOUSEMOVE",
    "MOUSEDOWN",
    "MOUSEUP",
    "MOUSEWHEEL",
    "MOUSEENTER",
    "MOUSELEAVE",
    "MOUSEDRAG",

    # Mouse buttons (with button info passed in event)
    "LEFT_DOWN",
    "LEFT_UP",
    "RIGHT_DOWN",
    "RIGHT_UP",
    "MIDDLE_DOWN",
    "MIDDLE_UP",

    # Keyboard
    "KEYDOWN",
    "KEYUP",
    "KEYPRESS",  # For character-producing keys
    "TEXTINPUT",  # Unicode / text events

    # Modifier keys
    "SHIFT_DOWN",
    "SHIFT_UP",
    "CTRL_DOWN",
    "CTRL_UP",
    "ALT_DOWN",
    "ALT_UP",
    "META_DOWN",
    "META_UP",

    # Scroll / touch / gestures
    "SCROLL",
    "TOUCHSTART",
    "TOUCHMOVE",
    "TOUCHEND",
    "TOUCHCANCEL",
    "PINCH",
    "ZOOM",

    # Drag and drop
    "DRAGENTER",
    "DRAGOVER",
    "DRAGLEAVE",
    "DROP",

    # App/system level
    "TIMER",
    "IDLE",
    "REDRAW",
    "EXPOSE", # X11

    # Custom / user-defined
    "USER_EVENT",
    "CUSTOM_EVENT",

    # Platform-specific extras
    "INPUTLANGCHANGE",
    "DISPLAYCHANGE",
    "POWERBROADCAST",
    "IME_STARTCOMPOSITION",
    "IME_ENDCOMPOSITION",
]

PIonEvent_Types = PIenum()
PIonEvent_Types.create_enum(_pionevent_types_str, 0)

class PIonEvent:
    """Pheonix Ion Event Class"""
    def __init__(self, typ: Union[int, str], **kwargs) -> None:
        self.type: int = 0

        if isinstance(typ, str):
            self.type = PIonEvent_Types.access(typ)
        else:
            self.type = typ

        self.data = kwargs

    def __repr__(self):
        return f"<PIonEvent type={self.type}, data={self.data}>"


"""Pyrion is a full custom GPU, yes you heard me right! a custom GPU. It is under development though!

Shader Script - Cindar Shader Language (CSL)
"""

__version__ = "0.1.0"
__author__ = "AkshuDev"
__name__ = "PyrionGPU"
__shader_language__ = "Cindar"
__shader_info__ = {
    "syntax": "Arris",
    "example": '@inc "stdio.alib"\n\nshader void CSL(in vector3 myvar) {\nreturn null;\n}'
}




"""DO NOT USE YET!

PENV stands for Pheonix Environment, it is a full virtual machine requiring no dependencies, it is capable of loading multiple OS made in either .py (python) or .vasm (Virtual Assembly). (NOTE: Learn more about .vasm from PVCpu)

It has a BIOS, it is capable of loading a 512-byte or less .vasm/.py file with AA55 signature. If it is a .py file it is run in complete restriction and can call the BIOS via -
```python
__int__(code:int)
```

PENV also has GUI capabilites and Driver support. (NOTE: Currently remapping the entire FileSystem so please wait till it is ready!)

PENV provides AOS:PENV by default, it is an OS that uses CLI and can run .aosf (AOS Executable File) and .caosf (Compressed AOS Executable File).

You can upgrade to AOS++:PENV for free but it will use more resources. It will provide GUI support and even Extension support.
"""

import os
import sys
import subprocess

base_path = os.path.join(os.path.dirname(os.path.abspath(__file__)))
module_path = os.path.join(base_path, "..", "..")

if module_path not in sys.path:
    sys.path.insert(0, module_path)

from phardwareitk.Memory import Memory as memory
from phardwareitk.PENV.shared import *
from phardwareitk.Extensions.C import *
from phardwareitk.Extensions.C.stdlib import *
from phardwareitk.Extensions.C.stdio import *
from phardwareitk.Extensions.C.stdint import *

win32, posix, unknown, os_ = get_os()

PBFS_DISK = os.path.join(base_path, "PheonixSSD.pbfs")

def force_os(_os: str, posix_based_os: bool = False) -> None:
    """Forces a OS so that the script will follow that specific os

    Args:
            _os (str): The OS you want to force.
    """
    _os = _os.lower()

    if _os == "windows":
        win32 = True
        posix = False
        unknown = False
    elif _os in posix_os:
        posix = True
        unknown = False
        win32 = False
    else:
        os_ = _os
        unknown = False
        win32 = False
        posix = False
        if posix_based_os:
            posix = True

def copy_folder_to_pbfs(pbfs, folder_path: str, reserved_size: int = 0, permissions: tuple = (1,1,1,1,0,0,1,0)):
    """
    Recursively copies a folder into PBFS.
    
    Args:
        pbfs (PBFS): The mounted PBFS instance.
        folder_path (str): Path to the folder to copy.
        reserved_size (int): Extra space to reserve per file.
        permissions (tuple): Permissions to apply to all files.
    """
    
    # Normalize folder path
    folder_path = os.path.abspath(folder_path)
    base_name = os.path.basename(folder_path)
    
    for root, dirs, files in os.walk(folder_path):
        # Compute relative path inside the drive
        rel_root = os.path.relpath(root, folder_path)
        if rel_root == ".":
            rel_root = ""  # root folder itself
        
        # Create folder entries (optional, for hierarchical tree)
        for d in dirs:
            folder_rel_path = os.path.join(base_name, rel_root, d).replace("\\", "/")
            pbfs.make_meta_entry(folder_rel_path + "/", 0, 0, 0, dir_=True)
        
        # Copy files
        for f in files:
            file_rel_path = os.path.join(base_name, rel_root, f).replace("\\", "/")
            abs_file_path = os.path.join(root, f)
            
            # Read file content
            with open(abs_file_path, "rb") as fd:
                content = fd.read()
            
            # Write file into PBFS
            pbfs.write_file(content, file_rel_path, reserved_size=reserved_size, permissions=permissions)

def start_penv(
    max_ram_bytes: int = 2 * 1000000,
    process_ram_size: int = 1 * 1000000,
    command_py: str = "python",
    bheight: int = 500,
    bwidth: int = 800,
    bdepth: int = 3,
    total_blocks:int = 2048,
    block_size: int = 512,
    disk_name: str = "PheonixSSD",
    include_uefi: bool = True,
    format_drive: bool = False,
    os: str = "" # Uses current OS
) -> None:
    """Starts Pheonix Virtual Environment"""
    from phardwareitk.PENV import PBFS
    from phardwareitk.PENV import bios

    memsize_part = process_ram_size // 4

    initialize(memsize_part)
    PBFS.initialize_pbfs(memsize_part)

    cmem = get_memory()

    if PBFS.validate_disk(PBFS_DISK, block_size) == False or format_drive:
        print("Creating PBFS Disk...")
        err = PBFS.format_disk(PBFS_DISK, total_blocks, block_size, b"disk_name")
        if err < 0:
            sys.exit(err)
    
    fs = PBFS.PBFS(PBFS_DISK, block_size, total_blocks)
    
    if not os == "":
        print("Formatting Drive...")
        PBFS.format_disk(PBFS_DISK, total_blocks, block_size, b"disk_name")
        print("Downloading OS...")
        copy_folder_to_pbfs(fs, os)
        print("Downloaded OS...")
        sys.exit(0)
    
if __name__ == "__main__":
    start_penv()



"""Base Class for GPU drivers"""

from typing import Any

class BaseGPUD:
    """Abstract GPU backend interface"""

    def init(self, display: Any, window: Any, create_and_attach_ctx: bool=False):
        """Initialize GPU context for the specified window"""
        raise NotImplementedError
    
    def clear(self, r: int, g: int, b: int, a: int):
        """Clear screen to a color."""
        raise NotImplementedError

    def swap(self):
        """Swap framebuffers / present frame."""
        raise NotImplementedError

    def shutdown(self):
        """Release context and resources."""
        raise NotImplementedError

    def load_function(self, name: str):
        """Get a raw pointer to a GPU function (e.g. glDrawArrays)."""
        raise NotImplementedError

    def create_context(self, display: Any, window: Any):
        """Returns the Created context in form of PIonContext"""
        raise NotImplementedError
    
    def destroy_context(self):
        """Destroys the created context"""
        raise NotImplementedError

    def viewport(self, w: int=800, h: int=600):
        """Changes/Makes the viewport"""
        raise NotImplementedError

    def create_shader(self, shader_type: str, source: str):
        """Creates a shader"""
        raise NotImplementedError

    def create_program(self, vertex_shader: Any, fragment_shader: Any):
        """Creates a program based on the shader's provided"""
        raise NotImplementedError

    def use_program(self, program: Any):
        """Uses a creates program"""
        raise NotImplementedError

    def create_buffer(self):
        """Creates a GPU Buffer"""
        raise NotImplementedError

    def bind_array_buffer(self, buf: Any):
        """Binds a Array buffer to the GPU"""
        raise NotImplementedError
    
    def buffer_data(self, data: Any, usage: Any):
        """Creats and initializes the buffer's data"""
        raise NotImplementedError

    def create_texture(self):
        """Creates a texture"""
        raise NotImplementedError

    def bind_texture(self, texture: Any, target: Any):
        """Binds the created texture to the GPU"""
        raise NotImplementedError

    def tex_image_2d(self, width: int, height: int, data: Any, target: Any, internal_format: Any, format_: Any, type_: Any):
        """Texture Image 2D"""
        raise NotImplementedError

    def create_framebuffer(self):
        """Creates a framebuffer"""
        raise NotImplementedError

    def bind_framebuffer(self, fbo: Any, target: Any):
        """Binds the created framebuffer to the GPU"""
        raise NotImplementedError
    
    def set_uniform_mat4(self, program: Any, name: str, mat: Any):
        """Set the value of a mat4 variable in a shader"""
        raise NotImplementedError
    
    def check_error(self):
        """Checks for error"""
        raise NotImplementedError


import os
import sys
import time
import json
import platform
import threading
import gc

os_files: list = [__file__]

path = sys.argv[1]
if not path in sys.path:
    sys.path.append(path)

from Memory import *

_mem = Memory(int(sys.argv[2]), int(sys.argv[3]))  # type: ignore
_mem.sys_block = MemBlock(64, os_files, "AOS-BLOCK-KERNEL", 0)  # type: ignore

_penv_path = sys.argv[5]
_drive_info = sys.argv[6]
_sys_info = json.loads(sys.argv[7])

if not _penv_path in sys.path:
    sys.path.append(_penv_path)

from PBFS import PBFS_DriveData

from AOS_FHD import setup_dir_tree

_file_manager = PBFS_DriveData()
_cd_ = "home"  # Current directory
_cd_backup = "home"
_user = "admin"
_pass = None

_file_manager.reload() # Reload the file manager

_cd_, _user = setup_dir_tree(_file_manager, _cd_, _user)

_file_manager.reload()

print("Kernel Loaded in Protected Mode!")

base_mode = "cli"

x: int = 0
y: int = 0

cmd = ""

def _p_sys_info() -> None:
    print("CPU cores:", str(_sys_info["cpu_cores"]))
    print("CPU Arch:", str(_sys_info["cpu_architecture"]))
    print("CPU Name:", str(_sys_info["processor_name"]))
    print("Host Name:", str(_sys_info["host_name"]))
    print("Host System:", str(_sys_info["host_system"]))
    print("Host Version:", str(_sys_info["host_version"]))
    print("Host Release:", str(_sys_info["host_release"]))
    print("Host Platform:", str(_sys_info["host_platform"]))
    print("Host Uname:", str(_sys_info["host_uname"]))
    print("Host Ram Size (GB):", str(_sys_info["ram_size"]))
    print("Host Storage Size (GB):", str(_sys_info["storage_size"]))

if _pass is not None:
    in_ = input("Password for User", _user, ": ")
    while True:
        if not in_ == _pass:
            print("Sorry wrong password, try again or type either shutdown or poweroff")
        else:
            print("Welcome!")
            break

        in_ = input("Password: ")

while True:
    cmd = input(f"{_cd_}\n|\n ---$: ").lower()
    cmd = cmd.split(" ")

    if cmd[0] == "poweroff" or cmd[0] == "shutdown":
        print("Shutting down the system...")
        break

    elif cmd[0] == "clear" or cmd[0] == "cls":
        os.system('cls' if platform.system().lower() == 'windows' else 'clear')

    elif cmd[0] == "wipe-current-user":
        sure = input("Are you sure you want to wipe the current user? (yes/no): ").lower()
        if sure == "yes":
            debug = False
            if len(cmd) > 1:
                if "-d:t" == cmd[1]:
                    debug = True

            _file_manager.reload()  # Reload the file manager to ensure we have the latest state
            print(f"Wiping current user data [{_file_manager.join_path(b'home', b'usrs', _user.encode('utf-8')).decode('utf-8')}]...")
            out = _file_manager.rm_dir(_file_manager.join_path(b"home", b"usrs", _user.encode("utf-8")), debug)
            if out != 0:
                print("Error wiping default user data:", out)
                continue
            print("Current user data wiped successfully!")
            print("Shutting Down...")
            break

    elif cmd[0] == "aos":
        # Get Param 1 as the command
        if not len(cmd) > 1:
            print("Usage: aos <command> <params>, use --help incase you don't know commands")
            continue

        cmd_aos = cmd[1]

        if cmd_aos == "--help":
            from aoshelp import cmd_aos_help
            cmd_aos_help(True, True)
            continue

        if cmd_aos == "wipe-drive":
            input_ = input("Are you sure, you want to erase everything? (Y/N): ")
            input_ = input_.lower()

            if not input_ == "y":
                continue

            from PBFS import start, PBFS_DriveData, PBFS_SuperBlock, init_vdrive
            init_vdrive(True)
            print("Drive is wiped, shutting down...")
            break

        elif cmd_aos == "cd-usr":
            if not len(cmd) > 2:
                print("Usage: aos cd-usr <user>")
                continue

            usr = cmd[2]

            user = _file_manager.join_path(b"home", b"usrs", usr.encode("utf-8")).decode("utf-8")

            if not _file_manager.dir_exists(user.encode("utf-8")):
                print("No such user exists!")

            out = _file_manager.write_file(b"home\\aosduser.aos", json.dumps({"user": f"home\\usrs\\{usr}_conf.aos"}).encode("utf-8"), [0x0, 0x5, (os_files, None)])
            if not out == 0:
                print("aos: ERROR, PBFS -", out)
                continue

        elif cmd_aos == "mk-usr":
            if not len(cmd) > 2:
                print("Usage: aos mk-usr <username> <OPTIONS>")
                continue

            usr = cmd[3]
            password = None

            if len(cmd) > 3:
                for option in cmd:
                    op = option.split(":")
                    if op[0] == "-p":
                        password = op[1]

            if password is None:
                password = 0xFFFF

            if "/\\=+{}'\";:,<>?!@#$%^&*()" in usr:
                print("Only alphabets, numbers and [-, _, .] are allowed in usernames")
                continue

            out = _file_manager.write_file(_file_manager.join_path(b"home", b"usrs", usr.encode("utf-8") + b"_conf.aos"), json.dumps({"name": usr, "password": password}).encode("utf-16"), [0x0, 0x5, os_files])

            if not out == 0:
                print("aos: ERROR, PBFS -", out)
                continue

        elif cmd_aos == "rm-usr":
            if not len(cmd) > 2:
                print("Usage aos rm-usr <user>")
                continue

            usr = cmd[3]
            user = _file_manager.join_path(b"home", b"usrs", usr.encode("utf-8"))

            if not _file_manager.dir_exists(user):
                print("No such user!")
                continue

            out = _file_manager.remove_file(user + b"_config.aos")

            if not out == 0:
                print("aos: ERROR, PBFS -", out)
                continue

            cmd = input("Verbose? (y/n): ").lower()

            if not cmd == "y":
                cmd = False
            else:
                cmd = True

            out = _file_manager.rm_dir(user, cmd)

            if not out == 0:
                print("aos: ERROR, PBFS -", out)
                continue

    elif cmd[0] == "sysinfo":
        print("System Information:")
        _p_sys_info()
    elif cmd[0] == "lsd":
        dir_ = None
        if len(cmd) > 1:
            dir_ = cmd[1]
        else:
            dir_ = input("Directory to list files in (leave empty for cd): ")
            if dir_ == "":
                dir_ = None

        if dir_ is None or dir_ == "":
            dir_ = _cd_

        if dir_.startswith("/") or dir_.startswith("\\"):
            dir_ = _file_manager.join_path(_cd_.encode("utf-8"), dir_.encode("utf-8")).decode("utf-8")

        print(f"Files/SubDirs in the directory [{dir_ if dir_ is not None else _cd_}]:")
        print(*_file_manager.list_dir(dir_), sep="\n")
    elif cmd[0] == "chf":
        src = None
        dst = None
        if len(cmd) > 2:
            src = cmd[1]
            dst = cmd[2]
        elif len(cmd) == 2:
            src = cmd[1]
            dst = input("Destination path in the file system: ")
        else:
            src = input("Source path on host system: ")
            dst = input("Destination path in the file system: ")

        data:bytes = b""

        if not os.path.exists(src):
            print("Source path does not exist!")
            continue
        else:
            with open(src, "rb") as f:
                data = f.read()

        dst = _file_manager.join_path(_cd_.encode("utf-8"), dst.encode("utf-8")).decode("utf-8")

        print("Copying file from host system to file system...")

        if isinstance(str, data):
            data.encode("utf-8")

        out = _file_manager.write_file(dst, data)
        if out != 0:
            print("Error copying file:", out)
        else:
            print("File copied successfully!")
    elif cmd[0] == "mvf":
        src = None
        dst = None
        if len(cmd) > 2:
            src = cmd[1]
            dst = cmd[2]
        elif len(cmd) == 2:
            src = cmd[1]
            dst = input("Destination path: ")
        else:
            src = input("Source path: ")
            dst = input("Destination path: ")

        src = _file_manager.join_path(_cd_.encode("utf-8"), src.encode("utf-8")).decode("utf-8")
        dst = _file_manager.join_path(_cd_.encode("utf-8"), dst.encode("utf-8")).decode("utf-8")

        print("Moving file from", src, "to", dst + "...")

        file_data = _file_manager.get_file(src)
        out = _file_manager.remove_file(src)
        if out != 0:
            print("Error removing file:", out)
        else:
            out = _file_manager.write_file(dst, file_data)
            if out != 0:
                print("Error writing file:", out)
            else:
                print("File moved successfully!")
    elif cmd[0] == "rmf":
        path = None
        if len(cmd) > 1:
            path = cmd[1]
        else:
            path = input("Path to remove: ")

        path = _file_manager.join_path(_cd_.encode("utf-8"), path.encode("utf-8")).decode("utf-8")

        print("Removing file at", path + "...")
        out = _file_manager.remove_file(path)
        if out != 0:
            print("Error removing file:", out)
        else:
            print("File removed successfully!")
    elif cmd[0] == "mkf":
        path:str = None
        data:str = None

        if len(cmd) > 2:
            path = cmd[1]
            data = _file_manager.get_file(cmd[2])
            if data == None:
                print("Error getting file data from path:", cmd[2])
                continue
        elif len(cmd) == 2:
            path = cmd[1]
        else:
            path = input("Path to create file: ")
            data = input("Data to write to the file (leave empty for no data): ")

        path = _file_manager.join_path(_cd_.encode("utf-8"), path.encode("utf-8")).decode("utf-8")

        print("Creating file at", path + "...")
        out = _file_manager.write_file(path.encode("utf-8"), data.encode("utf-8") if data else b"")
        if out != 0:
            print("Error creating file:", out)
        else:
            print("File created successfully!")
    elif cmd[0] == "gfc":
        path = ""
        if len(cmd) > 1:
            path = cmd[1]
        else:
            path = input("Path to get file from: ")

        path = _file_manager.join_path(_cd_.encode("utf-8"), path.encode("utf-8")).decode("utf-8")

        print("Getting file at", path + "...")
        file_data = _file_manager.get_file(path.encode("utf-8"))

        if not file_data:
            print("Error Getting file:", path)
            continue

        if isinstance(file_data, str):
            print("Error getting file:", file_data)
        else:
            print("File data:", file_data.decode("utf-8"))
    elif cmd[0] == "mkd":
        path = None
        if len(cmd) > 1:
            path = cmd[1]
        else:
            path = input("Path to create directory: ")

        path = _file_manager.join_path(_cd_.encode("utf-8"), path.encode("utf-8")).decode("utf-8")

        print("Creating directory at", path + "...")
        out = _file_manager.mk_dir(path.encode("utf-8"))
        if out != 0:
            print("Error creating directory:", out)
        else:
            print("Directory created successfully!")
    elif cmd[0] == "mvd":
        src = None
        dst = None
        if len(cmd) > 2:
            src = cmd[1]
            dst = cmd[2]
        elif len(cmd) == 2:
            src = cmd[1]
            dst = input("Destination path: ")
        else:
            src = input("Source path: ")
            dst = input("Destination path (the dir might be renamed, please include dir name at the end of the path): ")

        src = _file_manager.join_path(_cd_.encode("utf-8"), src.encode("utf-8")).decode("utf-8")
        dst = _file_manager.join_path(_cd_.encode("utf-8"), dst.encode("utf-8")).decode("utf-8")

        print("Moving directory from", src, "to", dst + "...")

        out = _file_manager.mv_dir(src.encode("utf-8"), dst.encode("utf-8"))
        if out != 0:
            print("Error moving directory:", out)
        else:
            print("Directory moved successfully!")
    elif cmd[0] == "rmd":
        path = None
        debug:bool = False

        if len(cmd) > 1:
            if "-d" in cmd[1]:
                debug = True if cmd[1] == "-d:t" else False
            else:
                path = cmd[1]
        elif len(cmd) > 2:
            path = cmd[1]
            debug = True if cmd[2] == "-d:t" else False
        else:
            path = input("Path to remove directory: ")

        path = _file_manager.join_path(_cd_.encode("utf-8"), path.encode("utf-8")).decode("utf-8")

        print("Removing directory at", path + "...")
        out = _file_manager.rm_dir(path.encode("utf-8"), debug)
        if out != 0:
            print("Error removing directory:", out)
        else:
            print("Directory removed successfully!")
    elif cmd[0] == "cd":
        if len(cmd) > 1:
            new_dir = cmd[1]
        else:
            new_dir = input("Enter directory to change to: ")

        if new_dir == "..":
                new_dir = new_dir.replace("..", "")
                if _cd_ != "home":
                    _cd_ = _cd_backup
                    if not _cd_backup == "home":
                        _cd_backup = _cd_backup.split("\\")
                        _cd_backup = _cd_backup[len(_cd_backup) - 2]
                    print("Changed directory to:", _cd_)
                else:
                    print("You are already in the home directory, cannot go up further.")
                    continue
        elif new_dir.startswith("./") or new_dir.startswith(".\\"):
            new_dir = new_dir.strip("./\\")
            new_dir = _file_manager.join_path(_cd_.encode("utf-8"), new_dir.encode("utf-8")).decode("utf-8")
            new_dir = new_dir.strip("./\\")
            new_dir_l = new_dir.replace("/", "\\").split("\\")
            cd_t = _cd_
            cd_backupt = _cd_backup
            work = True

            for part in new_dir_l:
                if part == "..":
                    cd_t = cd_backupt
                    if not cd_backupt == "home":
                        cd_backupL = cd_backupt.split("\\")
                        cd_backupL.pop(len(cd_backupL) - 1)
                        cd_backupt = "".join(*cd_backupL)
                else:
                    _cd_backup_t = _cd_
                    _cd_t = _file_manager.join_path(cd_t.encode("utf-8"), part.encode("utf-8")).decode("utf-8")
                    out = _file_manager.dir_exists(_cd_t.encode("utf-8"))
                    if out == False:
                        print("Directory doesn't exist:", _cd_t.encode("utf-8"))
                        work = False
                    else:
                        cd_t = _cd_t
                        cd_backupt = _cd_backup_t

            if work:
                _cd_ = cd_t
                _cd_backup = cd_backupt
                continue
        elif new_dir.startswith("~/") or new_dir.startswith("~\\"):
            new_dir = new_dir.strip("~/\\")
            new_dir = _file_manager.join_path(_cd_.encode("utf-8"), new_dir.encode("utf-8")).decode("utf-8")
            new_dir = new_dir.strip("~/\\")
            new_dir_l = new_dir.replace("/", "\\").split("\\")
            cd_t = "home"
            cd_backupt = "home"
            work = True

            for part in new_dir_l:
                if part == "..":
                    cd_t = cd_backupt
                    if not cd_backupt == "home":
                        cd_backupL = cd_backupt.split("\\")
                        cd_backupL.pop(len(cd_backupL) - 1)
                        cd_backupt = "".join(*cd_backupL)
                else:
                    _cd_backup_t = _cd_
                    _cd_t = _file_manager.join_path(cd_t.encode("utf-8"), part.encode("utf-8")).decode("utf-8")
                    out = _file_manager.dir_exists(_cd_t.encode("utf-8"))
                    if out == False:
                        print("Directory doesn't exist:", _cd_t.encode("utf-8"))
                        work = False
                    else:
                        cd_t = _cd_t
                        cd_backupt = _cd_backup_t

            if work:
                _cd_ = cd_t
                _cd_backup = cd_backupt
                continue
        else:
            new_dir = new_dir.strip("./\\")
            new_dir = _file_manager.join_path(_cd_.encode("utf-8"), new_dir.encode("utf-8")).decode("utf-8")
            new_dir = new_dir.strip("./\\")
            new_dir_l = new_dir.replace("/", "\\").split("\\")
            cd_t = _cd_
            cd_backupt = _cd_backup
            work = True

            for part in new_dir_l:
                if part == "..":
                    cd_t = cd_backupt
                    if not cd_backupt == "home":
                        cd_backupL = cd_backupt.split("\\")
                        cd_backupL.pop(len(cd_backupL) - 1)
                        cd_backupt = "".join(*cd_backupL)
                else:
                    _cd_backup_t = _cd_
                    _cd_t = _file_manager.join_path(cd_t.encode("utf-8"), part.encode("utf-8")).decode("utf-8")
                    out = _file_manager.dir_exists(_cd_t.encode("utf-8"))
                    if out == False:
                        print("Directory doesn't exist:", _cd_t.encode("utf-8"))
                        work = False
                    else:
                        cd_t = _cd_t
                        cd_backupt = _cd_backup_t

            if work:
                _cd_ = cd_t
                _cd_backup = cd_backupt
                continue

    elif cmd[0] == "tree-fs":
        print("Directory Tree (A-Z):")
        print("\nFiles:")
        print(*sorted(_file_manager.files), sep="\n")
        print("\nDirectories:")
        print(*sorted(_file_manager.dirs), sep="\n")
    elif cmd[0] == "reload-fs":
        print("Reloading file manager...")
        _file_manager.reload()
        print("File manager reloaded successfully!")

    else:
        print("Unknown command:", cmd[0])



"""Many Prebuilt and Custom Errors (You can create one using this also)"""

import os

import sys

PHardwareITK = len(sys.path) - 1
PHardwareITK_P = os.path.join(os.path.dirname(__file__), "..", "..")


import json

def setup_dir_tree(_file_manager:object, _cd_:str, _user:str) -> tuple[str, str]:
    _file_manager.reload()

    usrs = _file_manager.join_path(b"home", b"usrs")
    user = _file_manager.join_path(usrs, _user.encode("utf-8"))
    user_config = user + b"_config.aos"

    aosdusr = _file_manager.join_path(b"home", b"aosdusr.aos")

    out = _file_manager.get_file(aosdusr)

    if out is not None:
        # User found
        _user = json.loads(out.decode("utf-8"))["user"]
        _user = _user.replace("_config.aos", "")
        _cd_ = _user
        _user = _user.split("\\")
        _user = _user[len(_user) - 1]
        return (_cd_, _user)

    utemp = input("Sorry this OS user data seems to not be present, please specify the user you want to log into\n: ")

    user = _file_manager.join_path(b"home", b"usrs", utemp.encode("utf-8"))

    user_config = user + b"_config.aos"

    out = _file_manager.get_file(user_config)

    if out is not None:
        # User found
        return (user, user)
    else:
        print("No Such user found! Creating user", _user)
        # Create user admin
        user = _file_manager.join_path(usrs, _user.encode("utf-8"))
        user_config = user + b"_config.aos"
        user_config_data = {
        "name": _user,
        "default": True,
        "password": None
        }

        aosdusr_data = {
            "user": user_config.decode("utf-8")
        }

        out = _file_manager.write_file(user_config, json.dumps(user_config_data).encode("utf-16"), [0x0, 0x5, 0x6, [None]])

        if not out == 0:
            print("Error,", out)
            return None

        out = _file_manager.write_file(aosdusr, json.dumps(aosdusr_data).encode("utf-8"), [0x0, 0x5, 0x6, [None]])

        if not out == 0:
            print("Error,", out)
            return None

        out = _file_manager.mk_dir(user)

        if not out == 0:
            print("Error,", out)
            return None

        out = _file_manager.mk_dir(_file_manager.join_path(user, b"Downloads"))

        if not out == 0:
            print("Error,", out)
            return None

        out = _file_manager.mk_dir(_file_manager.join_path(user, b"Desktop"))

        if not out == 0:
            print("Error,", out)
            return None

        out = _file_manager.mk_dir(_file_manager.join_path(user, b"Documents"))

        if not out == 0:
            print("Error,", out)
            return None

        print("Done!")
        return user.decode("utf-8"), _user



"""OpenGL Driver implementation using BaseGPUD class"""

import ctypes
from ctypes import *
import sys
from typing import Any, Optional

from phardwareitk.GUI.pheonix_ion import PIonContext

from ._base import BaseGPUD
import platform

GL_COLOR_BUFFER_BIT = 0x00004000
GL_COMPILE_STATUS = 0x8B81
GL_LINK_STATUS = 0x8B82
GL_INFO_LOG_LENGTH = 0x8B84
GL_ARRAY_BUFFER = 0x8892
GL_STATIC_DRAW = 0x88E4
GL_TEXTURE_2D = 0x0DE1
GL_RGBA = 0x1908
GL_UNSIGNED_BYTE = 0x1401
GL_FRAMEBUFFER = 0x8D40
GL_VERTEX_SHADER = 0x8B31
GL_FRAGMENT_SHADER = 0x8B30

# Pixel format flags for Windows
PFD_DRAW_TO_WINDOW = 0x00000004
PFD_SUPPORT_OPENGL = 0x00000020
PFD_DOUBLEBUFFER = 0x00000001
PFD_TYPE_RGBA = 0
PFD_MAIN_PLANE = 0

FuncType = CFUNCTYPE
if sys.platform.startswith("win"):
    from ctypes import WINFUNCTYPE
    FuncType = WINFUNCTYPE

class UnsupportedPlatform(Exception):
    def __init__(self, message="Unsupported platform for window creation"):
        super().__init__(message)

class GLFunctions:
    """Wrapper for OpenGL functions — dynamically loads and binds symbols."""
    def __init__(self, driver):
        self._driver = driver
        self._lib = driver.lib

    def _get_proc_address(self, name: str, restype=None, argtypes: list=[]):
        """Use the driver's loader to resolve function addresses."""
        return self._driver.load_function(name, restype, argtypes)

    def _bind(self, name: str, restype=None, argtypes: list=[]):
        """Bind a GL function by name with argtypes/restype."""
        fn = self._get_proc_address(name, restype, argtypes)
        if not fn:
            raise RuntimeError(f"Failed to load GL function: {name}")
        
        setattr(self, name, fn)
        return fn

    def _load_gl_functions_common(self):
        """Load core GL 2.0+ functions required by GLDriver"""

        # Tell me which looks better as titles -
        # This is a title
        # ==== OR ====
        # --- This is a title ---

        # Shaders
        self.glCreateShader = self._bind("glCreateShader", c_uint, [c_uint])
        self.glShaderSource = self._bind("glShaderSource", None, [c_uint, c_int, POINTER(c_char_p), POINTER(c_int)])
        self.glCompileShader = self._bind("glCompileShader", None, [c_uint])
        self.glGetShaderiv = self._bind("glGetShaderiv", None, [c_uint, c_uint, POINTER(c_int)])
        self.glGetShaderInfoLog = self._bind("glGetShaderInfoLog", None, [c_uint, c_int, POINTER(c_int), POINTER(c_char)])
        self.glCreateProgram = self._bind("glCreateProgram", c_uint, [])
        self.glAttachShader = self._bind("glAttachShader", None, [c_uint, c_uint])
        self.glLinkProgram = self._bind("glLinkProgram", None, [c_uint])
        self.glGetProgramiv = self._bind("glGetProgramiv", None, [c_uint, c_uint, POINTER(c_int)])
        self.glGetProgramInfoLog = self._bind("glGetProgramInfoLog", None, [c_uint, c_int, POINTER(c_int), POINTER(c_char)])
        self.glUseProgram = self._bind("glUseProgram", None, [c_uint])

        # Buffers
        self.glGenBuffers = self._bind("glGenBuffers", None, [c_int, POINTER(c_uint)])
        self.glBindBuffer = self._bind("glBindBuffer", None, [c_uint, c_uint])
        self.glBufferData = self._bind("glBufferData", None, [c_uint, c_size_t, c_void_p, c_uint])

        # Textures
        self.glGenTextures = self._bind("glGenTextures", None, [c_int, POINTER(c_uint)])
        self.glBindTexture = self._bind("glBindTexture", None, [c_uint, c_uint])
        self.glTexImage2D = self._bind("glTexImage2D", None, [
            c_uint, c_int, c_int, c_int, c_int, c_int, c_uint, c_uint, c_void_p
        ])

        # Framebuffers
        self.glGenFramebuffers = self._bind("glGenFramebuffers", None, [c_int, POINTER(c_uint)])
        self.glBindFramebuffer = self._bind("glBindFramebuffer", None, [c_uint, c_uint])

        # Uniforms
        self.glGetUniformLocation = self._bind("glGetUniformLocation", c_int, [c_uint, c_char_p])
        self.glUniformMatrix4fv = self._bind("glUniformMatrix4fv", None, [c_int, c_int, c_uint, POINTER(c_float)])

        # Viewport + Errors
        self.glViewport = self._bind("glViewport", None, [c_int, c_int, c_int, c_int])
        self.glGetError = self._bind("glGetError", c_uint, [])

        # Color
        self.glClearColor = self._bind("glClearColor", None, [c_float, c_float, c_float, c_float])
        self.glClear = self._bind("glClear", None, [c_uint])

class GLDriver(BaseGPUD):
    """Cross Platform OpenGL Driver"""
    def __init__(self, libGL: Optional[str]=None, screen: Optional[int]=None):
        """Initialization
        
        Args:
            libGL (Optional, str): The path to the OpenGL Library, if None this class automatically tries to get it based on OS. Defaults to None
            screen (Optional, c_int): The screen to use, if None this class tries to get it based on the OS, Defaults to None."""
        self.display = None
        self.window = None
        self.ctx = None
        self.platform = platform.system().lower()
        self.lib_path = ""
        self.win32 = None
        self.user32 = None
        self.gdi32 = None
        self.hdc = None

        self.lib = None
        self.libGL = GLFunctions(self)
        if self.platform == "windows":
            from ctypes import wintypes as win
            self.win32 = win

        if libGL is not None:
            self.lib = cdll.LoadLibrary(libGL)
            self.lib_path = libGL
        else:
            if self.platform == "linux":
                self.lib = cdll.LoadLibrary("libGL.so.1")
                self.lib_path = "libGL.so.1"
            elif self.platform == "freebsd":
                self.lib = cdll.LoadLibrary("libGL.so")
                self.lib_path = "libGL.so"
                self.platform = "linux"
            elif self.platform == "openbsd":
                self.lib = cdll.LoadLibrary("libGL.so")
                self.lib_path = "libGL.so"
                self.platform = "linux"
            elif self.platform == "netbsd":
                self.lib = cdll.LoadLibrary("libGL.so")
                self.lib_path = "libGL.so"
                self.platform = "linux"
            elif self.platform == "windows":
                self.lib = ctypes.windll.opengl32
                self.lib_path = "opengl32.dll"
            elif self.platform == "darwin":
                self.lib = cdll.LoadLibrary("/System/Library/Frameworks/OpenGL.framework/OpenGL")
                self.lib_path = "/System/Library/Frameworks/OpenGL.framework/OpenGL"
            else:
                raise UnsupportedPlatform(f"Unsupported OS for GLX/OpenGL: {self.platform}, try passing lib path manually?")
        self.screen = screen

    def init(self, display: Any, window: Any, create_and_attach_ctx: bool=False):
        self._load_gl_functions()
        if self.platform == "linux":
            self.display = display
            self.window = window

            if create_and_attach_ctx:
                return self.create_context(display, window)
                
        elif self.platform == "windows":
            if not window:
                raise Exception("Provided Window's HWND is invalid, are you trying to draw on desktop? if yes, override this function!")

            self.window = window # display is useless

            self.gdi32 = ctypes.windll.gdi32
            self.user32 = ctypes.windll.user32

            if create_and_attach_ctx:
                return self.create_context(display, window)

    def clear(self, r: int, g: int, b: int, a: int):
        r = r / 255.0
        g = g / 255.0
        b = b / 255.0
        a = a / 255.0
        self.libGL.glClearColor(r, g, b, a)
        self.check_gl_error("glClearColor")
        self.libGL.glClear(GL_COLOR_BUFFER_BIT)
        self.check_gl_error("glClear")
    
    def swap(self):
        if self.platform == "linux":
            self.lib.glXSwapBuffers(self.display, self.window)
        elif self.platform == "windows":
            self.gdi32.SwapBuffers(self.hdc)

    def shutdown(self):
        self.destroy_context()

    def load_function(self, name: str, restype=None, argtypes: list=[]) -> c_void_p:
        addr = None
        if self.platform == "linux":
            addr = cast(self.lib.glXGetProcAddress(name.encode()), c_void_p)
        elif self.platform == "windows":
            addr = ctypes.cast(self.lib.wglGetProcAddress(name.encode()), ctypes.c_void_p)
            if not addr:
                try:
                    addr = getattr(self.lib, name)
                except AttributeError:
                    addr = None
        
            if not isinstance(addr, c_void_p): # it is a CFUNCTYPE
                addr = c_void_p(ctypes.cast(addr, c_void_p).value)

        if not addr:
            raise RuntimeError(f"Function {name} not found in GL driver")
        
        if isinstance(addr, c_void_p):
            addr = cast(addr, FuncType(restype, *argtypes))
        
        return addr
    
    def destroy_context(self):
        if self.ctx:
            if self.platform == "linux":
                self.lib.glXDestroyContext(self.display, self.ctx)
            elif self.platform == "windows":
                if self.ctx:
                    self.lib.wglDeleteContext(self.ctx)
                    self.ctx = None
                if self.hdc:
                    self.user32.ReleaseDC(self.ctx, self.hdc)
                    self.hdc = None
    
    def create_context(self, display, window):
        if self.platform == "linux":
            if self.ctx:
                self.destroy_context()

            attribs = (c_int * 5)(0x8010, 1, 0x8011, 1, 0)
            visual = self.lib.glXChooseVisual(display, 0, attribs)
            self.ctx = self.lib.glXCreateContext(display, visual, None, True)

            if not self.ctx:
                raise RuntimeError("Failed to create GLX Context")
            
            self.libGL._load_gl_functions_common()

            self.lib.glXMakeCurrent(display, window, self.ctx)

            return PIonContext(
                api_name="OpenGL",
                native_handle={
                    "display": display,
                    "window": window,
                    "glx_context": self.ctx,
                    "visual": visual
                },
                lib_path=self.lib_path,
                metadata={
                    "type": "GLX"
                }
            )
        elif self.platform == "windows":
            self.user32.ShowWindow(window, 1)
            self.user32.UpdateWindow(window)
            self.user32.SetForegroundWindow(window)
            self.user32.BringWindowToTop(window)
            self.user32.RedrawWindow(window, None, None, 0x0001 | 0x0004 | 0x0080)  # RDW_INVALIDATE|UPDATENOW|FRAME

            # Sleep a tick to let composition settle (important on Windows 10/11 + Intel)
            import time
            time.sleep(0.1)

            self.hdc = self.user32.GetDC(window)
            
            if not self.hdc:
                raise RuntimeError(f"GetDC failed — invalid window handle: {window}")

            class PIXELFORMATDESCRIPTOR(ctypes.Structure):
                _fields_ = [
                    ("nSize", ctypes.c_ushort),
                    ("nVersion", ctypes.c_ushort),
                    ("dwFlags", ctypes.c_uint),
                    ("iPixelType", ctypes.c_ubyte),
                    ("cColorBits", ctypes.c_ubyte),
                    ("cRedBits", ctypes.c_ubyte),
                    ("cRedShift", ctypes.c_ubyte),
                    ("cGreenBits", ctypes.c_ubyte),
                    ("cGreenShift", ctypes.c_ubyte),
                    ("cBlueBits", ctypes.c_ubyte),
                    ("cBlueShift", ctypes.c_ubyte),
                    ("cAlphaBits", ctypes.c_ubyte),
                    ("cAlphaShift", ctypes.c_ubyte),
                    ("cAccumBits", ctypes.c_ubyte),
                    ("cAccumRedBits", ctypes.c_ubyte),
                    ("cAccumGreenBits", ctypes.c_ubyte),
                    ("cAccumBlueBits", ctypes.c_ubyte),
                    ("cAccumAlphaBits", ctypes.c_ubyte),
                    ("cDepthBits", ctypes.c_ubyte),
                    ("cStencilBits", ctypes.c_ubyte),
                    ("cAuxBuffers", ctypes.c_ubyte),
                    ("iLayerType", ctypes.c_ubyte),
                    ("bReserved", ctypes.c_ubyte),
                    ("dwLayerMask", ctypes.c_uint),
                    ("dwVisibleMask", ctypes.c_uint),
                    ("dwDamageMask", ctypes.c_uint)
                ]
            pfd = PIXELFORMATDESCRIPTOR()
            pfd.nSize = ctypes.sizeof(PIXELFORMATDESCRIPTOR)
            pfd.nVersion = 1
            pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER
            pfd.iPixelType = PFD_TYPE_RGBA
            pfd.cColorBits = 32
            pfd.cDepthBits = 24
            pfd.cStencilBits = 8
            pfd.iLayerType = PFD_MAIN_PLANE

            exisiting_fmt = self.gdi32.GetPixelFormat(self.hdc)
            fmt = exisiting_fmt

            if exisiting_fmt == 0:
                fmt = self.gdi32.ChoosePixelFormat(self.hdc, ctypes.byref(pfd))
                if fmt == 0 or not self.gdi32.DescribePixelFormat(self.hdc, fmt, ctypes.sizeof(pfd), ctypes.byref(pfd)):
                    raise RuntimeError(f"ChoosePixelFormat failed - invalid HWND or HDC\n\tERR: {ctypes.GetLastError()}")
                if not self.gdi32.SetPixelFormat(self.hdc, fmt, ctypes.byref(pfd)):
                    raise RuntimeError(f"SetPixelFormat failed - could not set pixel format\n\tERR: {ctypes.GetLastError()}")

            # Create OpenGL rendering context
            self.ctx = self.lib.wglCreateContext(self.hdc)
            if not self.ctx:
                raise RuntimeError(f"Failed to create WGL Context\n\tERR: {ctypes.GetLastError()}")
            res = self.lib.wglMakeCurrent(self.hdc, self.ctx)
            if not res:
                raise RuntimeError(f"Failed to make WGL context current\n\tERR: {ctypes.GetLastError()}")


            current_ctx = self.lib.wglGetCurrentContext()
            if current_ctx != self.ctx:
                raise RuntimeError("WGL context is not current!")
            self.libGL._load_gl_functions_common()

            return PIonContext(
                api_name="OpenGL",
                native_handle={
                    "hwnd": self.window,
                    "hdc": self.hdc,
                    "hglrc": self.ctx
                },
                lib_path="opengl32.dll",
                metadata={"type": "WGL"}
            )

    def check_gl_error(self, func: str="Unknown"):
        error = self.libGL.glGetError()
        if error != 0:
            print(f"OpenGL function [{func}] exited with an Error: {error}")

    def check_error(self):
        return self.check_gl_error("<UNSPECIFIED by USER>")

    def viewport(self, w: int=800, h: int=600):
        self.libGL.glViewport(0, 0, w, h)
        self.check_gl_error("glViewport")

    def create_shader(self, shader_type: int, source: str) -> int:
        shader = self.libGL.glCreateShader(shader_type)
        self.check_gl_error("glCreateShader")
        src = c_char_p(source.encode())
        length = c_int(len(source))
        self.libGL.glShaderSource(shader, 1, byref(src), byref(length))
        self.check_gl_error("glShaderSource")
        self.libGL.glCompileShader(shader)
        self.check_gl_error("glCompileShader")

        compiled = c_int()
        self.libGL.glGetShaderiv(shader, GL_COMPILE_STATUS, byref(compiled))
        self.check_gl_error("glGetShaderiv")
        if not compiled.value:
            log_len = c_int()
            self.libGL.glGetShaderiv(shader, GL_INFO_LOG_LENGTH, byref(log_len))
            self.check_gl_error("glGetShaderiv")
            log = create_string_buffer(log_len.value)
            self.libGL.glGetShaderInfoLog(shader, log_len, None, log)
            self.check_gl_error("glGetShaderInfoLog")
            raise RuntimeError(f"Shader Compile Error: {log.value.decode()}")
        return shader
    
    def create_program(self, vertex_shader: int, fragment_shader: int) -> int:
        program = self.libGL.glCreateProgram()
        self.check_gl_error("glCreateProgram")
        self.libGL.glAttachShader(program, vertex_shader)
        self.check_gl_error("glAttachShader")
        self.libGL.glAttachShader(program, fragment_shader)
        self.check_gl_error("glAttachShader")
        self.libGL.glLinkProgram(program)
        self.check_gl_error("glLinkProgram")

        linked = c_int()
        self.libGL.glGetProgramiv(program, GL_LINK_STATUS, byref(linked))
        self.check_gl_error("glGetProgramiv")
        if not linked.value:
            log_len = c_int()
            self.libGL.glGetProgramiv(program, 0x8B84, byref(log_len))
            self.check_gl_error("glGetProgramiv")
            log = create_string_buffer(log_len.value)
            self.libGL.glGetProgramInfoLog(program, log_len, None, log)
            self.check_gl_error("glGetProgramInfoLog")
            raise RuntimeError(f"Program link error: {log.value.decode()}")
        return program
    
    def use_program(self, program: int):
        self.libGL.glUseProgram(program)
        self.check_gl_error("glUseProgram")

    def create_buffer(self) -> int:
        buf = c_uint()
        self.libGL.glGenBuffers(1, byref(buf))
        self.check_gl_error("glGenBuffers")
        return buf.value
    
    def bind_array_buffer(self, buf: int):
        self.libGL.glBindBuffer(GL_ARRAY_BUFFER, buf)
        self.check_gl_error("glBindBuffer")

    def buffer_data(self, data, usage: int=GL_STATIC_DRAW):
        arr_type = c_float * len(data)
        self.libGL.glBufferData(GL_ARRAY_BUFFER, sizeof(arr_type), arr_type(*data), usage)
        self.check_gl_error("glBufferData")

    def create_texture(self) -> int:
        tex = c_uint()
        self.libGL.glGenTextures(1, byref(tex))
        self.check_gl_error("glGenTextures")
        return tex.value
    
    def bind_texture(self, texture: int, target: int=GL_TEXTURE_2D):
        self.libGL.glBindTexture(target, texture)
        self.check_gl_error("glBindTexture")
    
    def tex_image_2d(self, width: int, height: int, data: Any=None, target: int=GL_TEXTURE_2D, internal_format: int=0x1908, format_: int=0x1908, type_: int=0x1401):
        self.libGL.glTexImage2D(target, 0, internal_format, width, height, 0, format_, type_, data)
        self.check_gl_error("glTexImage2D")

    def create_framebuffer(self) -> int:
        fbo = c_uint()
        self.libGL.glGenFramebuffers(1, byref(fbo))
        self.check_gl_error("glGenFramebuffers")
        return fbo.value
    
    def bind_framebuffer(self, fbo: int, target: int=GL_FRAMEBUFFER):
        self.libGL.glBindFramebuffer(target, fbo)
        self.check_gl_error("glBindFramebuffer")

    def set_uniform_mat4(self, program: int, name: str, mat):
        loc = self.libGL.glGetUniformLocation(program, name.encode())
        self.check_gl_error("glGetUniformLocation")
        self.libGL.glUniformMatrix4fv(loc, 1, 0, (c_float * 16)(*mat))
        self.check_gl_error("glUniformMatrix4fv")

    def _load_gl_functions(self):
        """Loads OpenGL Functions (platform-based, .e.g GLX for linux, WGL for Windows)"""
        if self.platform == "linux":
            self.lib.glXChooseVisual.argtypes = [c_void_p, c_int, POINTER(c_int)]
            self.lib.glXChooseVisual.restype = c_void_p
            self.lib.glXCreateContext.argtypes = [c_void_p, c_void_p, c_void_p, c_int]
            self.lib.glXCreateContext.restype = c_void_p
            self.lib.glXMakeCurrent.argtypes = [c_void_p, c_ulong, c_void_p]
            self.lib.glXMakeCurrent.restype = c_int
            self.lib.glXSwapBuffers.argtypes = [c_void_p, c_uint64]
            self.lib.glXSwapBuffers.restype = c_void_p
            self.lib.glXDestroyContext.argtypes = [c_void_p, c_void_p]
            self.lib.glXDestroyContext.restype = c_void_p
            self.lib.glXGetProcAddress.argtypes = [c_char_p]
            self.lib.glXGetProcAddress.restype = c_void_p

        elif self.platform == "windows":
            self.lib.wglCreateContext.argtypes = [c_void_p]
            self.lib.wglCreateContext.restype = c_void_p
            self.lib.wglMakeCurrent.argtypes = [c_void_p, c_void_p]
            self.lib.wglMakeCurrent.restype = c_int
            self.lib.wglDeleteContext.argtypes = [c_void_p]
            self.lib.wglDeleteContext.restype = c_int
            self.lib.wglGetCurrentContext.argtypes = []
            self.lib.wglGetCurrentContext.restype = c_void_p



"C-stdlib. Python - C - Standard Library"

from phardwareitk.Extensions import C
from typing import Union, Optional, Callable
import sys
import os
import random

_free_list: list[tuple[int, int]] = []
_atexit_funcs: list[Callable] = []

# Memory
def malloc(size_: C.size_t) -> Optional[C.pointer[C.void]]:
    """Allocate size on heap"""
    global _free_list
    
    size = C.align(size_.value, 2)

    for i, (addr, fsize) in enumerate(_free_list):
        if fsize >= size:
            C._heap_registry[addr] = size
            if fsize > size:
                new_free = (addr + size, fsize - size)
                _free_list[i] = new_free
            else:
                _free_list.pop(i)
            return C.pointer(addr, C.void)

    if size + C._heap_nextalloc > C._heap_size:
        return None

    addr = C._heap_nextalloc
    C._heap_registry[addr] = size
    C._heap_nextalloc += size
    return C.pointer(addr, C.void)

def free(ptr: C.pointer) -> None:
    """Frees the allocated heap memory for later use"""
    global _free_list
    addr = ptr.ptr_addr

    if not addr in C._heap_registry:
        raise OSError(f"Unknown Address {hex(addr)} for freeing!")

    size = C._heap_registry[addr]
    del C._heap_registry[addr]
    _free_list.append((addr, size))
    # Merge adjacent free addr
    # Sort by address
    _free_list.sort(key=lambda x: x[0])
    merged = []
    last_addr, last_size = _free_list[0]

    for addr, size in _free_list[1:]:
        if addr == last_addr + last_size:
            # Adjacent → merge
            last_size += size
        else:
            merged.append((last_addr, last_size))
            last_addr, last_size = addr, size

    merged.append((last_addr, last_size))
    _free_list = merged

def realloc(ptr: C.pointer, new_size: C.size_t) -> Optional[C.pointer[C.void]]:
    """Reallocate on the heap"""
    free(ptr)
    return malloc(new_size)

def calloc(nmemb: C.size_t, size: C.size_t) -> Optional[C.pointer[C.void]]:
    """Allocate specified number of members of specified size"""
    total_size = C.size_t(nmemb * size)
    new_ptr = malloc(total_size)
    if new_ptr is not None:
        offset = new_ptr.ptr_addr
        for i in range(nmemb.value):
            C.write_mem(b"\x00"*size.value, size.value, offset)
            offset += size.value
    return new_ptr

EXIT_SUCCESS = 0
EXIT_FALUIRE = 1

NULL = 0 # ((void*)0)

# Process Control
def abort() -> None:
    """Halts the program directly"""
    os.abort()

def exit(status: C.Int) -> None:
    """Exits the program"""
    global _atexit_funcs
    for func in _atexit_funcs:
        func()
    sys.exit(status.value)

def _Exit(status: C.Int) -> None:
    """Like exit but with minimal cleanup and, it doesn't call atexit"""
    os._exit(status.value)

def atexit(func: Callable) -> None:
    """Runs the function provided on exit of program"""
    global _atexit_funcs
    _atexit_funcs.insert(0, func)

# System

def getenv(name: C.pointer[C.char]) -> C.pointer[C.char]:
    """Gets Environment Variable"""
    pyname = name.deref().decode("utf-8")
    val = os.getenv(pyname)
    if val is None:
        return C.pointer(NULL, C.void)
    return C.cstring(val)

def system(command: C.pointer[C.char]) -> C.Int:
    """Runs a system command"""
    out = os.system(command.deref().decode("utf-8"))
    return C.Int(out)

def abs_(j: C.Int) -> C.Int:
    return C.Int(abs(j.value))

def labs(j: C.Int) -> C.Int:
    return C.Int(abs(j.value))

def div_(numer: C.Int, denom: C.Int):
    class div_t:
        quot: int
        rem: int
        def __init__(self, q, r):
            self.quot, self.rem = q, r
    return div_t(numer.value // denom.value, numer.value % denom.value)

#  Conversion Routines

def atoi(nptr: C.pointer[C.char]) -> C.Int:
    s = nptr.deref().decode("utf-8")
    return C.Int(int(s.strip() or 0))

def atof(nptr: C.pointer[C.char]) -> C.double:
    s = nptr.deref().decode("utf-8")
    try:
        return C.double(float(s.strip()))
    except ValueError:
        return C.double(0.0)

def atol(nptr: C.pointer[C.char]) -> C.Int:
    s = nptr.deref().decode("utf-8")
    try:
        return C.Int(int(s.strip()))
    except ValueError:
        return C.Int(0)

# Random

def srand(seed: C.Int) -> None:
    random.seed(seed.value)

def rand() -> C.Int:
    return C.Int(random.randint(0, 32767))

# Search & Sort

def qsort(base: C.pointer[C.array], nmemb: C.size_t, size: C.size_t, compar: Callable): pass

def bsearch(key: C.pointer[C.char], base: C.pointer[C.array], nmemb: C.size_t, size: C.size_t, compar: Callable): pass

# Multibyte / Wide Char

def mbstowcs(pwcs, s, n): pass
def wcstombs(s, pwcs, n): pass
def mbtowc(pwc, s, n): pass
def wctomb(s, wc): pass
def mblen(s, n): pass



"""Hardware Hyper Game (HGame).

To Use, pass parameters in **HGame_Settings**. Also no need to save it, as it has a automated system, to provide all functions with every parameter needed. Check file for code. And then the stage is yours!
"""

import os
import sys
import platform

PHardwareITK = len(sys.path) - 1
PHardwareITK_P = os.path.join(os.path.dirname(__file__), "..", "..")

if not sys.path[PHardwareITK] == PHardwareITK_P:
    sys.path.append(PHardwareITK_P)

from phardwareitk.Extensions import *
from phardwareitk.Extensions import HyperOut as HOut

class HGame_OUT:
    """This class defined **Output** locations available for HGame.
    """
    @staticmethod
    def console() -> int:
        return 0

    @staticmethod
    def stdout() -> int:
        return HGame_OUT.console()

    @staticmethod
    def messagebox() -> int:
        return 2

    @staticmethod
    def file() -> int:
        return 4

    if platform.platform().lower() == "windows":
        def winEventManager() -> int:
            return 3

class HGame_Settings:
    """This class offers the settings for using HGame.

    Parameters:
        output [HGame_OUT]: Defined where errors and logs will be displayed. Defaults to **HGame_OUT.stdout**
        debug [bool]: If true, Debug mode will be used. Defaults to False.
        filesystem_encrypt [bool]: If true, Creation of files via HGame_FileManager will be encrypted. Defaults to False.
    """
    _instances:list = []

    def __init__(self, output:HGame_OUT=HGame_OUT.stdout, debug:bool=False, filesystem_encrypt:bool=False) -> None:
        if not isinstance(output, HGame_OUT):
            HOut.printH("[HGAME ERROR] ->\n\tArgument [output] takes a [HGame_OUT] class only.", FontEnabled=True, Font=TextFont(font_color=Color("red")))
            raise TypeError("See Above for ERROR.")
        if not isinstance(debug, bool):
            HOut.printH("[HGAME ERROR] ->\n\tArgument [debug] takes a [bool] class only.", FontEnabled=True, Font=TextFont(font_color=Color("red")))
            raise TypeError("See Above for ERROR.")
        if not isinstance(filesystem_encrypt, bool):
            HOut.printH("[HGAME ERROR] ->\n\tArgument [filesystem_encrypt] takes a [bool] class only.", FontEnabled=True, Font=TextFont(font_color=Color("red")))
            raise TypeError("See Above for ERROR.")

        self.output:int = output()
        self.debug:bool = debug
        self.filesystem_encrypt:bool = filesystem_encrypt

        if not HGame_Settings._instances:
            HGame_Settings._instances.append(self)

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(HGame_Settings, cls).__new__(cls)
            cls._instance.initialize_values() # call the initialization function
        return cls._instance

    @property
    def output_(self) -> int:
        return self.output

    @property
    def debug_(self) -> bool:
        return self.debug

    @property
    def filesystem_encrypt_(self) -> bool:
        return self.filesystem_encrypt

    @classmethod
    def get_first_instance(cls):
        if cls._instances:
            return cls._instances[0]
        else:
            return None

    def __str__(self) -> str:
        return f"Hyper Game Settings"

    def __repr__(self) -> str:
        return f"Hyper Game Settings ->\n\tOutput: {self.output}\n\tDebug: {self.debug}\n\tFile Encrypt: {self.filesystem_encrypt}"

from phardwareitk.GUI import gui

# phardwareitk/HGame/core.py

class GameObject:
    """
    Represents an entity within the game world.

    Attributes:
        name (str): The name of the GameObject.
        position (tuple): The (x, y) coordinates of the GameObject.
        components (list): A list of Component objects attached to the GameObject.
    """

    def __init__(self, name="GameObject", position=(0, 0)):
        """
        Initializes a new GameObject.

        Args:
            name (str, optional): The name of the GameObject. Defaults to "GameObject".
            position (tuple, optional): The (x, y) coordinates. Defaults to (0, 0).
        """
        self.name = name
        self.position = position
        self.components = []

    def add_component(self, component):
        """
        Adds a Component to the GameObject.

        Args:
            component (Component): The Component object to add.
        """
        self.components.append(component)
        component.game_object = self

    def get_component(self, component_type):
        """
        Retrieves a Component of a specific type.

        Args:
            component_type (type): The type of Component to retrieve.

        Returns:
            Component or None: The Component object if found, otherwise None.
        """
        for component in self.components:
            if isinstance(component, component_type):
                return component
        return None

class Component:
    """
    Base class for components that can be attached to GameObjects.

    Attributes:
        game_object (GameObject or None): The GameObject this Component is attached to.
    """

    def __init__(self):
        """
        Initializes a new Component.
        """
        self.game_object = None

class Sprite(Component):
    """
    Represents a visual sprite for a GameObject.

    Attributes:
        image (str): The path to the image file.
    """

    def __init__(self, image):
        """
        Initializes a new Sprite.

        Args:
            image (str): The path to the image file.
        """
        super().__init__()
        self.image = image

class Transform(Component):
    """
    Represents the position, rotation, and scale of a GameObject.

    Attributes:
        position (tuple): The (x, y) coordinates.
        rotation (float): The rotation angle in degrees.
        scale (tuple): The (x, y) scale.
    """

    def __init__(self, position=(0, 0), rotation=0.0, scale=(1.0, 1.0)):
        """
        Initializes a new Transform.

        Args:
            position (tuple, optional): The (x, y) coordinates. Defaults to (0, 0).
            rotation (float, optional): The rotation angle. Defaults to 0.0.
            scale (tuple, optional): The (x, y) scale. Defaults to (1.0, 1.0).
        """
        super().__init__()
        self.position = position
        self.rotation = rotation
        self.scale = scale

class Script(Component):
    """
    Base class for user-defined scripts attached to GameObjects.

    Users should override the Start, Update, and End methods.
    """

    def start(self):
        """
        Called when the GameObject is created.
        """
        pass

    def update(self):
        """
        Called every frame.
        """
        pass

    def end(self):
        """
        Called when the GameObject is destroyed.
        """
        pass

class Input:
    """
    Handles user input.

    Class Methods:
        get_key(key): Returns True if the specified key is pressed.
        get_mouse_position(): Returns the current mouse position.
    """

    @staticmethod
    def get_key(key):
        """
        Checks if a key is currently pressed.

        Args:
            key (str): The key to check (e.g., "W", "SPACE", "LEFT").

        Returns:
            bool: True if the key is pressed, False otherwise.
        """
        # (Replace with your actual input handling logic)
        return False  # Placeholder

    @staticmethod
    def get_mouse_position():
        """
        Returns the current mouse position.

        Returns:
            tuple: The (x, y) coordinates of the mouse.
        """
        # (Replace with your actual input handling logic)
        return (0, 0)  # Placeholder

class Scene:
    """
    Represents a game scene containing GameObjects.

    Attributes:
        game_objects (list): A list of GameObjects in the scene.
    """

    def __init__(self):
        """
        Initializes a new Scene.
        """
        self.game_objects = []

    def add_game_object(self, game_object):
        """
        Adds a GameObject to the scene.

        Args:
            game_object (GameObject): The GameObject to add.
        """
        self.game_objects.append(game_object)

    def remove_game_object(self, game_object):
        """
        Removes a GameObject from the scene.

        Args:
            game_object (GameObject): The GameObject to remove.
        """
        if game_object in self.game_objects:
            self.game_objects.remove(game_object)
            



"""This file includes all classes to write Basic 'C' Code inside Python without the need of Cython. (OLD VERSION)"""

from typing import *
import sys
import errno
import os

MODULE_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..")

if not MODULE_DIR in sys.path:
    sys.path.append(MODULE_DIR)

# We need memory to work
from Memory import Memory as mem

# Basic Macros
NULL = 0
UNINITIALIZED: bytes = b"\xff"

FALSE = 0
TRUE = 1

EXIT_SUCCESS = 0
EXIT_FALUIRE = 1

METADATA_MAGIC: bytes = b"\xF7\xFF\xF7\xFF"

base = 0x1

heap_base = base
heap_ptr = heap_base

next_alloc = heap_base

size = 64
stack_base = size
stack_ptr = stack_base

memory = mem.Memory(64, 0, None, False, 0)  # Defaults to 64 bytes as size

FREE_ADDR = {}


def reset_mem(size_: int, base_: int = 0x0, debug: bool = False) -> None:
    """Resets the memory but updates its size"""
    global size
    global memory
    global base
    global next_alloc
    global stack_base
    global stack_ptr
    global heap_base
    global heap_ptr
    
    if base_ != 0x0:
        base = base_
        next_alloc = base
    size = size_
    
    stack_base = size_
    stack_ptr = stack_base
    heap_base = base_
    heap_ptr = heap_base
    
    memory = mem.Memory(size, 0, None, debug, 0)
    FREE_ADDR.clear()


def get_next_alloc() -> int:
    global next_alloc
    return next_alloc


def set_next_alloc(value: int) -> None:
    global next_alloc
    next_alloc = value


def append_next_alloc(value: int) -> None:
    global next_alloc
    global heap_ptr
    next_alloc += value + 8  # 8 bytes for metadata
    heap_ptr += value + 8


def align(
    address: int, alignment: int
) -> int:  # To align to a specific address in memory
    if address % alignment == 0:
        return address
    return address + (alignment - (address % alignment))


def full_delete() -> None:
    """Deletes everything from memory"""
    global size
    reset_mem(size)


def del_mem(addr: int, size_: int, stack: bool = True) -> None:
    """NULLS/frees the specified memory block for the specified size"""
    global memory
    global size

    if addr > size:
        raise OSError(
            "Trying to access memory outside of the virtual memory space. [ERROR:PHardwareITK:Extensions:C - del_mem]"
        )

    memory.write_ram(b"\x00", addr, size_)
    if not stack:
        FREE_ADDR[addr] = size_


def get_memory() -> mem.Memory:
    global memory
    return memory


def set_mem(addr: int, size_: int, data: bytes) -> None:
    """Sets memory to specfied data"""
    global size
    global memory

    if addr > size:
        raise OSError(
            f"Trying to access memory outside of the virtual memory space using address [{hex(addr)}], while virtual memory has a size of [{hex(size)}]. [ERROR:PHardwareITK:Extensions:C - set_mem]"
        )

    memory.write_ram(METADATA_MAGIC + size_.to_bytes(4, "little"), addr - 8, 0)
    memory.write_ram(data, addr, 0)

    if addr in list(FREE_ADDR.keys()):
        FREE_ADDR.pop(addr)

def write_mem(addr: int, size_: int, data: bytes) -> None:
    """Writes data to specified address, unlike set_mem, this doesnt create metadata"""
    global memory
    global size
    
    if addr > size:
        raise OSError(
            f"Trying to access memory outside of the virtual memory space using address [{hex(addr)}], while virtual memory has a size of [{hex(size)}]. [ERROR:PHardwareITK:Extensions:C - write_mem]"
        )
    
    memory.write_ram(data, addr, 0)

def get_mem(addr: int, size_: int) -> bytes:
    """Gets memory"""
    global memory
    global size

    if addr > size:
        raise OSError(
            "Trying to access memory outside of the virtual memory space. [ERROR:PHardwareITK:Extensions:C - get_mem]"
        )

    return memory.get_ram(size_, addr)

def push_stack(data: bytes, size: int) -> None:
    """Pushes data onto stack"""
    global stack_ptr
    stack_ptr -= align(size, 2) + 8
    data = data.ljust(align(size, 2), b"\x00")
    set_mem(stack_ptr + 8, size, data)
    
def pop_stack(size:int) -> bytes:
    """Pops data from the stack and returns it"""
    global stack_ptr
    
    metadata = get_mem(stack_ptr, 8)
    if not metadata[0:4] == METADATA_MAGIC:
        raise OSError(f"Invalid Metadata in stack at [{hex(stack_ptr)}], {metadata}")

    stack_ptr += 8
    data = get_mem(stack_ptr, size)
    stack_ptr += size
    return data

def get_mem_size() -> int:
    global size
    return size


def metadata_verify(data: bytes) -> int:
    if not data == METADATA_MAGIC:
        return 1

    return 0


def get_size_metadata(addr: int) -> int:
    data: bytes = get_mem(addr, 8)
    if len(data) < 8:
        raise OSError(
            f"Not enough data to read metadata, expected 8 bytes, got less - {data}"
        )

    magic = data[0:4]
    if not metadata_verify(magic) == 0:
        raise OSError(f"Falsly mapped data in memory, {data}")

    return int.from_bytes(data[4:8], "little")


# Variables, and types
# The most basic needs
class Size_t:
    """size_t"""

    def __init__(self, value: int) -> None:
        self.value = value  # bytes
        self.bytes = value
        self.bits = value * 8
        self.kb = value / 1024
        self.mb = self.kb / 1024
        self.gb = self.mb / 1024
        self.tb = self.gb / 1024

    def __repr__(self) -> str:
        return f"{self.bytes}"


class Uint8_t:
    size = 1

    """uint8_t"""

    def __init__(self, value: int = 0) -> None:
        self.og_value = value

        if isinstance(value, str):
            if not len(value) == 0:
                value = ord(value)
            else:
                value = 0
        elif isinstance(value, bytes):
            value = int.from_bytes(value, "big")

        if value > 0xFF:
            raise TypeError("Value exceeds 8 bits or 1 byte")
        elif value < 0x0:
            raise TypeError("Value less than 0")

        self.value = value
        self.data = value.to_bytes(1, "little")
        self.size = 1
        push_stack(self.data, self.size)
        self.address = stack_ptr + 8

        self.deleted = False

    def __repr__(self) -> str:
        return f"{self.og_value}"

    def __del__(self) -> None:
        if getattr(sys, "meta_path", None) is None:
            return  # Interpreter is shutting down; skip cleanup

        if self.deleted:
            return

        del_mem(self.address - 8, self.size + 8)

        self.deleted = True


class Uint16_t:
    size = 2

    """uint16_t"""

    def __init__(self, value: int = 0) -> None:
        self.og_value = value

        if isinstance(value, str):
            if not len(value) == 0:
                value = ord(value)
            else:
                value = 0
        elif isinstance(value, bytes):
            value = int.from_bytes(value, "big")

        if value > 0xFFFF:
            raise TypeError("Value Exceeds 16 bits or 2 bytes")
        elif value < 0x0:
            raise TypeError("Value less than 0")

        self.value = value
        self.data = value.to_bytes(2, "little")
        self.size = 2
        push_stack(self.data, self.size)
        self.address = stack_ptr + 8
        self.deleted = False

    def __repr__(self) -> str:
        return f"{self.og_value}"

    def __del__(self) -> None:
        if getattr(sys, "meta_path", None) is None:
            return  # Interpreter is shutting down; skip cleanup

        if self.deleted:
            return

        del_mem(self.address - 8, self.size + 8)

        self.deleted = True


class Uint32_t:
    size = 4

    """uint32_t"""

    def __init__(self, value: int = 0) -> None:
        self.og_value = value

        if isinstance(value, str):
            if not len(value) == 0:
                value = ord(value)
            else:
                value = 0
        elif isinstance(value, bytes):
            value = int.from_bytes(value, "big")

        if value > 0xFFFFFFFF:
            raise TypeError("Value Exceeds 32 bits or 4 bytes")
        elif value < 0x0:
            raise TypeError("Value less than 0")

        self.value = value
        self.data = value.to_bytes(4, "little")
        self.size = 4
        push_stack(self.data, self.size)
        self.address = stack_ptr + 8

        self.deleted = False

    def __repr__(self) -> str:
        return f"{self.og_value}"

    def __del__(self) -> None:
        if getattr(sys, "meta_path", None) is None:
            return  # Interpreter is shutting down; skip cleanup

        if self.deleted:
            return

        del_mem(self.address - 8, self.size + 8)

        self.deleted = True


class Uint64_t:
    size = 8

    """uint64_t"""

    def __init__(self, value: int = 0) -> None:
        self.og_value = value

        if isinstance(value, str):
            if not len(value) == 0:
                value = ord(value)
            else:
                value = 0
        elif isinstance(value, bytes):
            value = int.from_bytes(value, "big")

        if value > 0xFFFFFFFFFFFFFFFF:
            raise TypeError("Value Exceeds 64 bits or 8 bytes")
        elif value < 0x0:
            raise TypeError("Value less than 0")

        self.value = value
        self.data = value.to_bytes(8, "little")
        self.size = 8
        push_stack(self.data, self.size)
        self.address = stack_ptr + 8

        self.deleted = False

    def __repr__(self) -> str:
        return f"{self.og_value}"

    def __del__(self) -> None:
        if getattr(sys, "meta_path", None) is None:
            return  # Interpreter is shutting down; skip cleanup

        if self.deleted:
            return

        del_mem(self.address - 8, self.size + 8)

        self.deleted = True


class Int8:
    """int8"""
    size = 1

    def __init__(self, value: int = 0) -> None:
        self.og_value = value

        if isinstance(value, str):
            if not len(value) == 0:
                value = ord(value)
            else:
                value = 0
        elif isinstance(value, bytes):
            value = int.from_bytes(value, "big")

        if not (-0x80 <= value <= 0x7F):
            raise TypeError("Value Doesn't fit in 8 bits or 1 bytes")
        
        self.value = value
        self.data = value.to_bytes(1, "little", signed=True)
        self.size = 1
        push_stack(self.data, self.size)
        self.address = stack_ptr + 8

        self.deleted = False

    def __repr__(self) -> str:
        return f"{self.og_value}"

    def __del__(self) -> None:
        if self.deleted:
            return

        del_mem(self.address - 8, self.size + 8)

        self.deleted = True


class Int16:
    """int16"""
    size = 2

    def __init__(self, value: int = 0) -> None:
        self.og_value = value

        if isinstance(value, str):
            if not len(value) == 0:
                value = ord(value)
            else:
                value = 0
        elif isinstance(value, bytes):
            value = int.from_bytes(value, "big")

        if not (-0x8000 <= value <= 0x7FFF):
            raise TypeError("Value Doesn't fit in 16 bits or 2 bytes")
    
        self.value = value
        self.data = value.to_bytes(2, "little", signed=True)
        self.size = 2
        push_stack(self.data, self.size)
        self.address = stack_ptr + 8

        self.deleted = False

    def __repr__(self) -> str:
        return f"{self.og_value}"

    def __del__(self) -> None:
        if self.deleted:
            return

        del_mem(self.address - 8, self.size + 8)

        self.deleted = True


class Int32:
    """int32"""
    size = 4

    def __init__(self, value: int = 0) -> None:
        self.og_value = value

        if isinstance(value, str):
            if not len(value) == 0:
                value = ord(value)
            else:
                value = 0
        elif isinstance(value, bytes):
            value = int.from_bytes(value, "big")

        if not (-0x80000000 <= value <= 0x7FFFFFFF):
            raise TypeError(f"Value {value} Doesn't fit in 32 bits or 4 bytes")

        self.value = value
        self.data = value.to_bytes(4, "little", signed=True)
        self.size = 4
        push_stack(self.data, self.size)
        self.address = stack_ptr + 8

        self.deleted = False

    def __repr__(self) -> str:
        return f"{self.og_value}"

    def __del__(self) -> None:
        if self.deleted:
            return

        del_mem(self.address - 8, self.size + 8)

        self.deleted = True


class Int64:
    """int64"""
    size = 8

    def __init__(self, value: int = 0) -> None:
        self.og_value = value

        if isinstance(value, str):
            if not len(value) == 0:
                value = ord(value)
            else:
                value = 0
        elif isinstance(value, bytes):
            value = int.from_bytes(value, "big")

        if not (-0x8000000000000000 <= value <= 0x7FFFFFFFFFFFFFFF):
            raise TypeError("Value Doesn't fit in 64 bits or 8 bytes")
        
        self.value = value
        self.data = value.to_bytes(8, "little", signed=True)
        self.size = 8
        push_stack(self.data, self.size)
        self.address = stack_ptr + 8

        self.deleted = False

    def __repr__(self) -> str:
        return f"{self.og_value}"

    def __del__(self) -> None:
        if self.deleted:
            return

        del_mem(self.address - 8, self.size + 8)

        self.deleted = True


# Types of int
class Short(Int16):
    """short"""

    def __init__(self, value: int = 0) -> None:
        super().__init__(value)

    def __repr__(self) -> str:
        return super().__repr__()

    def __del__(self) -> None:
        super().__del__()


class Long(Int32):
    """long"""

    def __init__(self, value: int = 0) -> None:
        super().__init__(value)

    def __repr__(self) -> str:
        return f"Long of value [{self.value}] of size [{self.size}]"

    def __del__(self) -> None:
        super().__del__()


class Int(Int32):
    """int"""

    def __init__(self, value: int = 0) -> None:
        super().__init__(value)

    def __repr__(self) -> str:
        return super().__repr__()

    def __del__(self) -> None:
        super().__del__()


# Chars and string access
class Char(Int8):
    """char"""

    def __init__(self, value: Union[str, int, bytes] = "\0") -> None:
        super().__init__(value)

    def __repr__(self) -> str:
        return super().__repr__()

    def __del__(self) -> None:
        super().__del__()


# Pointers and void
class Void:
    """void"""

    def __init__(self) -> None:
        self.size = 0
        self.value = None
        self.address = NULL  # Void

    def __repr__(self) -> str:
        return "<void>"


class Pointer(Uint64_t):
    """*<value>"""

    def __init__(
        self, type_: object = Void, object_: object = Void(), save: bool = False
    ) -> None:
        self.type = type_
        self.pointer_address = object_

        if object_ == NULL:
            self.pointer_address = NULL

        if object_ is None:
            self.pointer_address = NULL

        if isinstance(object_, int):
            pass
        elif isinstance(object_, str):
            raise TypeError("Cannot create a pointer to a string (python)")
        elif isinstance(object_, bytes):
            raise TypeError("Cannot create a pointer to bytes (python)")
        elif object_ is not None:
            self.pointer_address = getattr(object_, "address", NULL)  # A C.py file object

        self.obj = None
        self.save = save

        if save:
            self.obj = object_
        
        super().__init__(self.pointer_address)
        # Don't Map * to memory as uint64_t does it

    def dereference(self) -> object:
        if isinstance(self.type, Void):
            raise TypeError(
                "Cannot dereference a void pointer without casting it to another type."
            )

        if self.save:
            if self.obj is not None:
                return self.obj

        size = get_size_metadata(self.pointer_address - 8)

        data = get_mem(self.pointer_address, size)

        return data

    def cast(self, type_: object, object_:object=None) -> None:
        """Casts a pointer to another type"""
        self.type = type_

    def __repr__(self) -> str:
        return f"{self.pointer_address.to_bytes(8, 'little').hex()}"

    def __del__(self) -> None:
        super().__del__()

    @classmethod
    def __class_getitem__(cls, type_: object):
        return Pointer # Just keep it simple


# String access


def get_string(ptr: Union[Pointer[Char], str, bytes]) -> str:
    """Returns a string from an	memory address"""
    if isinstance(ptr, str): 
        return ptr
    elif isinstance(ptr, bytes):
        return ptr.decode("utf-8")
    
    string = b""
    addr = ptr.pointer_address
    
    size = get_size_metadata(addr - 8)

    string = get_mem(addr, size).removesuffix(b"\x00\x00")

    return string.decode("utf-8")


CHAR_PTR = Pointer[Char]


# Arrays
class Array:
    """<type>[<size>]"""

    def __init__(self, type: object, size: int) -> None:
        self.type = type
        self.size = size * type.size
        self.asize = size

        self.tsize = type.size

        self.array = []

        self.deleted = False

    def fill(self, data: bytes) -> None:
        for i in range(0, self.asize, self.tsize):
            self.array.append(self.type(data[i : self.tsize + i]))

    def __repr__(self) -> str:
        string = ""

        for obj in self.array:
            string += obj.__repr__()

        return string

    def __del__(self) -> None:
        if getattr(sys, "meta_path", None) is None:
            return  # Interpreter is shutting down; skip cleanup

        if self.deleted:
            return
        for obj in self.array:
            obj.__del__()

        self.deleted = True

    @classmethod
    def __class_getitem__(cls, type_: object = None, size_: int = 0):
        return Array

# Funcs (Basic)
def sizeof(value: object) -> Size_t:
    if isinstance(value, int):
        return Size_t(4) # size of 32-bit int
    elif isinstance(value, str):
        return Size_t(len(value))
    elif isinstance(value, bytes):
        return Size_t(len(value))
    else:
        val = None
        try:
            val = Size_t(value.size)
        except AttributeError:
            val = Size_t(value().size)
        
        return val

# Funcs (Memory)
def malloc(size: Union[int, Size_t]) -> Pointer[Void]:
    """Allocate memory"""
    if isinstance(size, Size_t):
        size = size.bytes

    if not isinstance(size, int):
        raise TypeError("Size is not a Size_t or a int.")

    for addr, sz in FREE_ADDR.items():
        if sz >= size + 8:  # We found a free block
            FREE_ADDR.pop(
                addr
            )  # We don't set the memory because C malloc doesn't either.
            return Pointer(Void, addr + 8)

    # ELSE
    addr = next_alloc
    append_next_alloc(size)

    write_mem(
        addr, 8, METADATA_MAGIC + size.to_bytes(4, "little")
    )  # Set the metadata, we don't set the memory to unint because c doesnt either

    return Pointer(Void, addr + 8)


def free(ptr: Pointer, chunk_size:int=1024*16) -> int:
    """Free memory (doesn't delete pointer). This frees memory in chunks to be memory safe and default chunk size if 16KB"""
    addr = ptr.pointer_address
    size = get_mem_size()

    if addr > size:
        raise OSError(
            "Trying to access memory outside of the virtual memory space. [ERROR:PHardwareITK:Extensions:C - free]"
        )

    size = get_size_metadata(addr - 8)
    offset = 0
    remaining = size
    
    while remaining > 0:
        to_free = min(chunk_size, remaining)
        write_mem(addr + offset, to_free, UNINITIALIZED * to_free)
        remaining -= to_free
        offset += to_free

    FREE_ADDR[addr] = size

    return 0

def calloc(nmemb: Union[int, Size_t], size: Union[int, Size_t]) -> Pointer[Void]:
    """Allocate memory and set all values to 0"""
    if isinstance(nmemb, Size_t):
        nmemb = nmemb.bytes
    if isinstance(size, Size_t):
        size = size.bytes

    if not isinstance(nmemb, int):
        raise TypeError("Error nmemb is not a Size_t or a int")
    elif not isinstance(size, int):
        raise TypeError("Error size is not a Size_t or a int")

    total_size = nmemb * size

    return malloc(total_size)


def realloc(ptr: Pointer[Void], size: Union[int, Size_t]) -> Pointer[Void]:
    """Reallocate memory with new size"""
    # Free old memory
    free(ptr)

    # Allocate new memory
    return malloc(size)


def memcpy(
    dest: Pointer[Void], src: Pointer[Void], size: Union[int, Size_t]
) -> Pointer[Void]:
    """Copies [size] bytes from src to dest"""
    addr_dest = dest.pointer_address
    addr_src = src.pointer_address

    if isinstance(size, Size_t):
        size = size.bytes

    if not isinstance(size, int):
        raise TypeError("Argument size must be a Size_t or a int")

    data = get_mem(addr_src, size)
    write_mem(addr_dest, size, data)
    return Pointer(Void, addr_dest)


def memmove(
    dest: Pointer[Void], src: Pointer[Void], size: Union[int, Size_t]
) -> Pointer[Void]:
    """Moves [size] bytes from src to dest"""
    addr_src = src.pointer_address

    if isinstance(size, Size_t):
        size = size.bytes

    if not isinstance(size, int):
        raise TypeError("Argument size must be a Size_t or a int")

    memcpy(dest, src, size)
    write_mem(addr_src, size, UNINITIALIZED * size)
    
    del src

    return Pointer(Void, dest.pointer_address)


def memset(ptr: Pointer[Void], value: int, size: Union[int, Size_t]) -> Pointer[Void]:
    """Sets [size] number of bytes of memory pointed to by [ptr] to the byte value [value]"""
    if isinstance(size, Size_t):
        size = size.bytes

    if not isinstance(size, int):
        raise TypeError("Argument size must be a Size_t or a int")

    value = value.to_bytes(1, "little")
    val = value * size

    write_mem(ptr.pointer_address, size, val)

    return ptr


def memchr(ptr: Pointer[Void], value: int, size: Union[int, Size_t]) -> Pointer[Void]:
    """Finds the needed byte in memory"""
    addr = ptr.pointer_address
    cbyte = 0  # current addr
    searched_bytes = 0

    if isinstance(size, Size_t):
        size = size.bytes

    if not isinstance(size, int):
        raise TypeError("Argument size must be a Size_t or a int")

    data = get_mem(addr, size)

    value = value.to_bytes(1, "little")

    while True:
        if searched_bytes > size:
            return None

        if data[cbyte] == value:
            return Pointer(Void, addr + cbyte)


def memcmp(ptr1: Pointer[Void], ptr2: Pointer[Void], size: Union[int, Size_t]) -> int:
    """Compares two data in memory"""
    if isinstance(size, Size_t):
        size = size.bytes

    if not isinstance(size, int):
        raise TypeError("Argument size must be a Size_t or a int")

    addr1 = ptr1.pointer_address
    addr2 = ptr2.pointer_address
    data1 = get_mem(addr1, size)
    data2 = get_mem(addr2, size)

    if data1 == data2:
        return 0
    else:
        return 1


def write(ptr: Pointer[Void], data: bytes, size: Union[int, Size_t], no_meta: bool=False) -> int:
    """Writes data to memory"""
    if isinstance(size, Size_t):
        size = size.bytes

    if not isinstance(size, int):
        raise TypeError("Argument size must be a Size_t or a int")
        
    if no_meta == True:
        write_mem(ptr.pointer_address, size, data)
        return 0

    set_mem(ptr.pointer_address, size, data)

    return 0

def read(ptr: Pointer[Void], size: Union[int, Size_t]) -> bytes:
    """Reads data from memory"""
    if isinstance(size, Size_t):
        size = size.bytes

    if not isinstance(size, int):
        raise TypeError("Argument size must be a Size_t or a int")

    addr = ptr.pointer_address
    data = get_mem(addr, size)

    return data


# String creation
def make_string(value: str) -> Pointer[Char]:
    """Makes a char*"""

    if not isinstance(value, str):
        raise TypeError("Value must be a [str]")

    if not value.endswith("\x00\x00"):
        value += "\x00\x00"
    
    # Make a list of Char
    push_stack(value.encode("utf-8"), len(value))
    out = Pointer(Char, stack_ptr + 8)

    return out  # char* is basically the pointer to the first char in a string


# Structs
class Struct:
    """struct {...}"""

    def __init__(self, structure: dict) -> None:
        """Format:
        {
                '<Name>': {
                        'type': <type in form of one of the class here>,
                        'value': <value>,
                        'ptr_type': <type of pointer. NOTE: Only if your 'type' is a Pointer and this is Optional, it will default to Void>,
                        'array_type': <type of array. NOTE: Only if your 'type' is a Array>,
                        'array_size': <size of array. NOTE: Only if your 'type' is a Array>
                }
        }
        """
        self.structure = structure
        self.size = 0
        self.value = NULL

        self.data = []  # Data of the struct, Format = (<name>, <size>)

        self.get_size()  # Set the size

        self.address = align(next_alloc, self.size)
        append_next_alloc(self.size)

        set_mem(self.address, self.size, UNINITIALIZED * self.size)

        self.deleted = False

        for key, value in self.structure.items():
            self.data.append((key, sizeof(value["type"]).bytes))

    def access(self, name: str) -> Any:
        """Returns the value of the object"""
        val = None

        try:
            val = self.structure[name]["value"]
        except Exception:
            return -1

        return val

    def get_offset(self, name: str) -> int:
        """Returns the offset of the object"""
        offset = 0

        for key, value in self.data:
            if key == name:
                return offset

            offset += value + 8  # 8 bytes for metadata

    def set(self, name: str, value: object) -> int:
        """Sets the new value of a part of the struct. NOTE: The new value must be of the old defined type"""
        # C Doesn't check stuff like this (i mean the program not compiler)
        try:
            self.structure[name]["value"] = value
        except Exception:
            print("Struct Set failed: Assignment Failed")
            return -3

        return 0

    def get_size(self) -> int:
        """Gets the size of the struct"""
        self.size = 0
        for key, value in self.structure.items():
            self.size += sizeof(value["type"]).bytes

        return self.size

    def fill_b(self, data: bytes, byteorder: str = "big") -> int:
        """Fills the entire struct by the provided value (bytes)"""
        self.get_size()

        if len(data) < self.size:
            return -1  # Not enough data

        index = 0
        for name, field in self.structure.items():
            field_size = sizeof(field["type"]).bytes
            field_data = data[index : index + field_size]

            field_type = field["type"]
            field_ptr_type = field.get("ptr_type", Void)
            field_array_type = field.get("array_type", Int)
            field_array_size = field.get("array_size", 1)
            field_value = int.from_bytes(field_data, byteorder)

            if issubclass(field_type, Char):
                field_value = Char(field_data.decode("utf-8"))
            elif issubclass(field_type, Pointer):
                if issubclass(field_ptr_type, Char):
                    field_value = make_string(field_data.decode("utf-8"))
                elif issubclass(field_ptr_type, Array):
                    field_value = Array(field_array_type, field_array_size)
                    field_value.fill(field_data)
                    field_value = Pointer(Array, field_value)
                elif issubclass(field_ptr_type, Void):
                    # void* means it can point to anything
                    # We are assuming anything
                    field_value = Pointer(Void)
            elif issubclass(field_type, Array):
                field_value = Array(field_array_type, field_array_size)
                field_value.fill(field_data)  # Fill the array with the data
            elif issubclass(field_type, Void):
                # Void in C means anything/nothing
                # We are assuming anything
                field_value = Void()
            else:
                try:
                    field_value = field_type(field_value)
                except Exception as e:
                    print(f"Unknown type -> {field_type}\nException=>{e}")
                    return -1

            self.set(name, field_value)

            index += field_size

    def fill_f(self, file: TextIO, byteorder: str = "big") -> int:
        """Fills the struct from a file"""
        self.get_size()
        data = file.read(self.size)
        return self.fill_b(data, byteorder)

    def write_b(self, buffer_: Pointer[Void], max_memory_usage:int=1024*1024*64) -> int:
        """Writes the filled structure to any point in memory.
        
        Parameters:
            buffer_ (Pointer[Void]): The address in memory to write to
            max_memory_usage (int): The maximum memory usage by this function to prevent crashes to Python/Terminal/Environment. Default is 64 MB / 1024*1024*64 bytes.
            
        Returns:
            int: Return Code. 0 Means success.
        """
        if self.size < 1:
            return -1  # Size = 0

        addr = buffer_.pointer_address
        offset = 0
        self.get_size()
        
        MAX_WRITE = 1024 # 1KB
        total_written = 0
        
        for key, val in self.structure.items():
            try:
                value = val["value"]
                size = sizeof(value).bytes
            
                # NOTE: array are very complicated to work with and i will add support for arrays but not now!
                if isinstance(value, Pointer):
                    data = value.dereference() # Value stored in pointer
                    if value.save:
                        size = data.size
                    else:
                        size = len(data)
                        
                    if total_written + size > max_memory_usage:
                        print("[CRITICAL ERROR] Memory Usage Exceeded, Stopping immedietly: C.Struct.write_b")
                        sys.exit(errno.ENO_MEM)
                    
                    if size > MAX_WRITE:
                        for i in range(0, size, MAX_WRITE):
                            write(Pointer(Void, addr + offset + i), data[i:i + MAX_WRITE], min(MAX_WRITE, size-i), no_meta=True)
                    else:
                        write(Pointer(Void, addr + offset), data, size, no_meta=True)
                    offset += size
                    total_written += size
                elif isinstance(value, str):
                    if total_written + size > max_memory_usage:
                        print("[CRITICAL ERROR] Memory Usage Exceeded, Stopping immedietly: C.Struct.write_b")
                        sys.exit(errno.ENO_MEM)
                    
                    write(Pointer(Void, addr + offset), value.encode("utf-8"), size, no_meta=True)
                    
                    offset += size
                    total_written += size
                elif isinstance(value, bytes):
                    if total_written + size > max_memory_usage:
                        print("[CRITICAL ERROR] Memory Usage Exceeded, Stopping immedietly: C.Struct.write_b")
                        sys.exit(errno.ENO_MEM)
                    
                    write(Pointer(Void, addr + offset), value, size, no_meta=True)
                    
                    offset += size
                    total_written += size
                elif not isinstance(value, Array):
                    if total_written + size > max_memory_usage:
                        print("[CRITICAL ERROR] Memory Usage Exceeded, Stopping immedietly: C.Struct.write_b")
                        sys.exit(errno.ENO_MEM)
                    
                    write(Pointer(Void, addr + offset), read(Pointer(Void, value), size), size, no_meta=True)
                    offset += size
                    total_written += size
                else:
                    print("Array are not yet supported")
            except Exception as e:
                print(f"Exception occured on field [{key}] => {e}")
                return -1
                
        return 0

    def dereference(self, ptr: Pointer[Void]) -> None:
        """Dereferences a pointer to the struct and returns the struct object."""
        self.get_size()
        del_mem(self.address, self.size)
        self.address = ptr.pointer_address
        self.fill_b(get_mem(self.address, self.size))

    def __del__(self) -> None:
        if getattr(sys, "meta_path", None) is None:
            return  # Interpreter is shutting down; skip cleanup

        if self.deleted:
            return  # Already deleted

        del_mem(self.address, self.size)
        self.deleted = True

        # Free the objects
        for key, value in self.structure.items():
            val = value["value"]
            if not val is None:
                del val

    @classmethod
    def __class_getitem__(cls, struct: dict = {}, size_: int = 0):
        return (cls, struct, size_)


def dereference_struct(ptr: Pointer[Struct], structure: dict) -> Struct:
    """Dereferences a pointer to a struct and returns the struct object.

    Args:
            ptr (Pointer[Struct]): Pointer to the struct object
            structure (dict): The struct definition

    Returns:
            Struct: The struct object
    """
    return Struct(structure).dereference(ptr)



"""This is the new and improved C Extension, which allows for easy usage. It allows emulating C in python to do tasks that Cython/C Can do in python by emulating it."""

from typing import Union, Optional, Any
from phardwareitk.Memory.Memory import Memory as __memory__

_memory: Optional[__memory__] = None
_memsize: int = 64

_heap_ptr: int = 0x0
_heap_nextalloc: int = 0x0
_heap_registry: dict[int, int] = {}
_heap_ratio: float = 0.50
_heap_size: int = int(_memsize * _heap_ratio) 

_stack_ptr: int = _memsize
_stack_frames: list[int] = []

def align(v1: int, v2: int) -> int:
    return ((v1 + v2 - 1) // v2) * v2

def reset_mem(new_size: int) -> None:
    """Resets the memory with new size, expanding or shrinking it"""
    global _memory
    global _memsize
    global _stack_ptr
    global _heap_size
    _memsize = new_size
    _memory = __memory__(new_size, system_size=0, proc_sector_size=0, Block=None)
    _stack_ptr = new_size
    _heap_size = int(_memsize * _heap_ratio)

def initialize(size: int=64) -> None:
    """Initializes the memory, very important to use this before using anything else!"""
    reset_mem(size)

def get_memory() -> __memory__:
    """Returns the global memory"""
    global _memory
    return _memory

def write_mem(data:bytes, size:int, addr:int) -> None:
    """Writes data of specified size at the specified address"""
    global _memory
    global _memsize

    if addr > _memsize:
        raise OSError(
            f"Trying to access memory outside of the virtual memory space using address [{hex(addr)}], while virtual memory has a size of [{hex(_memsize)}]. [ERROR:PHardwareITK:Extensions:C - set_mem]"
        )

    data_ = data[:size]
    if size == len(data): data_ = data

    _memory.write_ram(data_, addr, 0)

def read_mem(size:int, addr: int) -> bytes:
    """Reads the specified size from the specified address from memory."""
    global _memory
    global _memsize

    if addr > _memsize:
        raise OSError(
            f"Trying to access memory outside of the virtual memory space using address [{hex(addr)}], while virtual memory has a size of [{hex(_memsize)}]. [ERROR:PHardwareITK:Extensions:C - set_mem]"
        )

    return _memory.get_ram(size, addr)

def delete_mem(size: int, addr: int) -> None:
    """Deletes/Clears memory at specified address for specified size"""
    write_mem(b"\x00"*size, size, addr)

def stack_push(data: bytes) -> None:
    """Pushes data onto the stack"""
    global _stack_ptr

    size = align(len(data), 8)
    if size > len(data):
        data.rjust(size, b"\x00")
    elif size < len(data):
        data = data[:size]

    write_mem(data, size, _stack_ptr)
    _stack_ptr -= size

def stack_pop(count: int) -> bytes:
    """Pops 8 bytes off the stack for specifed 'count' times and returns it concatinated"""
    global _stack_ptr
    global _memsize

    data: bytes = b""

    for i in range(0, count):
        _stack_ptr += 8
        if _stack_ptr > _memsize: raise ValueError("Nothing to pop, Exceeding memory limits!")
        data += read_mem(8, _stack_ptr)

    return data

def push_frame():
    """Starts a new stack frame"""
    global _stack_frames, _stack_ptr
    _stack_frames.append(_stack_ptr)

def pop_frame():
    """Ends a stack frame and restores stack pointer"""
    global _stack_frames, _stack_ptr
    if not _stack_frames:
        raise RuntimeError("No stack frame to pop!")
    _stack_ptr = _stack_frames.pop()

class CBaseType():
    size = 1
    signed = False
    def __init__(self, value:int) -> None:
        self.cls = type(self)
        if self.cls.signed:
            if self.cls.size == 1:
                if value < -0x80 or value > 0x7F:
                    raise ValueError("Value doesn't fit in 1 bytes!")
            elif self.cls.size == 2:
                if value < -0x8000 or value > 0x7FFF:
                    raise ValueError("Value doesn't fit in 2 bytes!")
            elif self.cls.size == 4:
                if value < -0x80000000 or value > 0x7FFFFFFF:
                    raise ValueError("Value doesn't fit in 4 bytes!")
            elif self.cls.size == 8:
                if value < -0x8000000000000000 or value > 0x7FFFFFFFFFFFFFFF:
                    raise ValueError("Value doesn't fit in 8 bytes!")
            else:
                raise ValueError(f"Unknown size: {self.cls.size} bytes")

        else:
            if value < 0:
                raise ValueError("Value is signed while it was supposed to be unsigned!")
            if self.cls.size == 1:
                if value > 0xFF:
                    raise ValueError("Value doesn't fit in 1 byte!")
            elif self.cls.size == 2:
                if value > 0xFFFF:
                    raise ValueError("Value doesn't fit in 2 bytes!")
            elif self.cls.size == 4:
                if value > 0xFFFFFFFF:
                    raise ValueError("Value doesn't fit in 4 bytes!")
            elif self.cls.size == 8:
                if value > 0xFFFFFFFFFFFFFFFF:
                    raise ValueError("Value doesn't fit in 8 bytes!")
            else:
                raise ValueError(f"Unknown size: {self.cls.size} bytes")

        self.value = value
        self.address = _stack_ptr
        stack_push(value.to_bytes(self.cls.size, "little", signed=self.cls.signed))

        self.deleted = False

    def __repr__(self) -> str:
        return str(self.value)

    def __sizeof__(self) -> int:
        return self.cls.size

    def __neg__(self) -> int:
        if not self.cls.signed: raise ValueError("Cannot negate unsigned!")
        return -self.value

    def __pos__(self) -> int:
        return self.value

    def __abs__(self) -> int:
        if self.value < 0: return self.value * -1
        return self.value

    def __invert__(self) -> int:
        return ~self.value

    def __add__(self, other: object) -> int:
        if isinstance(other, int): return self.value + other
        return self.value + getattr(other, "value", 0)

    def __sub__(self, other: object) -> int:
        if isinstance(other, int): return self.value - other
        return self.value - getattr(other, "value", 0)

    def __mul__(self, other: object) -> int:
        if isinstance(other, int): return self.value * other
        return self.value * getattr(other, "value", 0)

    def __truediv__(self, other: object) -> float:
        if isinstance(other, int): return self.value / other
        return self.value / getattr(other, "value", 0)
    
    def __floordiv__(self, other: object) -> int:
        if isinstance(other, int): return self.value // other
        return self.value // getattr(other, "value", 0)

    def __mod__(self, other: object) -> int:
        if isinstance(other, int): return self.value % other
        return self.value % getattr(other, "value", 0)

    def __pow__(self, other: object, modulo=None) -> int:
        if isinstance(other, int): return self.value ** other
        return self.value ** getattr(other, "value", 0)

    def __lshift__(self, other: object) -> int:
        if isinstance(other, int): return self.value << other
        return self.value << getattr(other, "value", 0)

    def __rshift__(self, other: object) -> int:
        if isinstance(other, int): return self.value >> other
        return self.value << getattr(other, "value", 0)

    def __and__(self, other: object) -> int:
        if isinstance(other, int): return self.value & other
        return self.value & getattr(other, "value", 0)

    def __or__(self, other: object) -> int:
        if isinstance(other, int): return self.value | other
        return self.value | getattr(other, "value", 0)

    def __xor__(self, other: object) -> int:
        if isinstance(other, int): return self.value ^ other
        return self.value ^ getattr(other, "value", 0)

    def __iadd__(self, other: object) -> int:
        val = 0
        if isinstance(other, int):
            val = other
        else:
            val = getattr(other, "value", 0)
        self.value = self.value + val
        write_mem(self.value.to_bytes(self.cls.size, "little", signed=self.cls.signed), self.cls.size, self.address)
        return self.value

    def __isub__(self, other: object) -> int:
        val = 0
        if isinstance(other, int):
            val = other
        else:
            val = getattr(other, "value", 0)
        self.value = self.value - val
        write_mem(self.value.to_bytes(self.cls.size, "little", signed=self.cls.signed), self.cls.size, self.address)
        return self.value

    def __index__(self) -> int:
        return self.value

    def __del__(self) -> None:
        if not self.deleted:
            self.deleted = True
        return # C ends the stack frame for the func but we cannot do that

class char(CBaseType):
    """A single byte"""
    size = 1
    signed = True

    def __init__(self, value: Union[str, int]) -> None:
        val = 0
        if isinstance(value, str):
            val = ord(value)
        else:
            val = value
        super().__init__(val)

    def __repr__(self) -> str:
        val = ""
        try:
            val = chr(self.value)
        except Exception:
            val = self.value
        return f"{val}"

class short(CBaseType):
    """2 bytes"""
    size = 2
    signed = True

class int_(CBaseType):
    """4 bytes"""
    size = 4
    signed = True

class long(CBaseType):
    """8 bytes"""
    size = 8
    signed = True

class unsigned_char(CBaseType):
    """unsigned char"""
    size = 1
    signed = False

class unsigned_short(CBaseType):
    """unsigned short"""
    size = 2
    signed = False

class unsigned_int(CBaseType):
    """unsigned int"""
    size = 4
    signed = False

class unsigned_long(CBaseType):
    """unsigned long"""
    size = 8
    signed = False

class double(CBaseType):
    """64-bit floating point"""
    size = 8
    signed = True

    def __init__(self, value: float) -> None:
        import struct
        self.value = float(value)
        self.address = _stack_ptr
        data = struct.pack("<d", self.value)
        stack_push(data)
        self.deleted = False

    def __repr__(self):
        return str(self.value)

    def __float__(self):
        return self.value

class void():
    """Just nothing, plain void"""
    def __init__(self) -> None:
        return # Incomplete type (Just a note, I mean like even in C it is incomplete, it means nothing!)

class pointer(CBaseType):
    """A 64-bit Pointer"""
    size = 8
    signed = True
    
    def __init__(self, obj: object, typ: type=void) -> None:
        self.typ = typ
        val = 0x0
        if isinstance(obj, int):
            val = obj
            if val < 0: raise ValueError("Address cannot be negative!")
        else:
            val = getattr(obj, "address", 0x0)
            if val == 0x0:
                print("[WARNING]: Object provided to pointer doesn't have any address! Defaulting to 0x0")

        super().__init__(val)
        self.ptr_addr = val # Another self var just for ease of use for the developer

    def deref(self) -> bytes:
        """Dereference the pointer. &pointer"""
        if self.typ == char:
            data = b""
            addr = self.ptr_addr
            while True:
                b_ = read_mem(1, addr)
                if b_ == b"\x00" or len(b_) == 0:
                    break
                data += b_
                addr += 1
            return data

        return read_mem(getattr(self.typ, "size", 8), self.value)

    def cast(self, newtype: type) -> None:
        """Cast the pointer to another type"""
        self.typ = newtype

    def __repr__(self) -> str:
        return f"{self.typ.__name__}* of <{hex(self.ptr_addr)}> at <{hex(self.address)}>"

    @classmethod
    def __class_getitem__(cls, *args):
        return pointer

# typedefs for ease-of-use
ptr = pointer
Pointer = pointer
Void = void
Char = char
Short = short
Int = int_
Long = long
UChar = unsigned_char
UShort = unsigned_short
UInt = unsigned_int
ULong = unsigned_long
size_t = unsigned_long
Size_t = unsigned_long
Double = double

class array():
    """C-Array, Example: 
    ``` c
        char myarray[5] // array of 5 characters/5 bytes
    ```
    """
    def __init__(self, typ: type, len: int, address: Optional[int]=None) -> None:
        self.typ = typ
        self.len = len
        self.element_size = getattr(typ, "size", 1)
        self.total_size = self.element_size * len
        self.address = 0
        self.deleted = True
        if not address:
            global _heap_nextalloc
            base_addr = _heap_nextalloc
            _heap_nextalloc += self.total_size
            _heap_registry[base_addr] = self.total_size
            self.address = base_addr
        
            self.deleted = False 
            write_mem(b"\x00"*self.total_size, self.total_size, self.address)
        else:
            self.address = address

    def __getitem__(self, index: int) -> CBaseType:
        if not (0 <= index < self.len):
            raise IndexError(f"Array index [{index}] is out of range! 0-{self.len} available.")

        addr: int = self.address + (index * self.element_size)
        raw: bytes = read_mem(self.element_size, addr)
        data: int = int.from_bytes(raw, "little", signed=getattr(self.typ, "signed", True))
        ret = self.typ.__new__(self.typ)
        ret.value = data
        ret.address = addr
        ret.deleted = True # Temp
        return ret

    def __setitem__(self, index: int, value: Union[int, CBaseType]) -> None:
        if not (0 <= index < self.len):
            raise IndexError(f"Array index [{index}] is out of range! 0-{self.len} available.")

        val = value
        if isinstance(value, int) or isinstance(value, self.typ):
            if not isinstance(value, int):
                val = getattr(value, "value", 0)
            else:
                val = value
        else:
            raise ValueError(f"Trying to set array index [{index}] with a value of neither [{self.typ}] or [int] types!")

        addr: int = self.address + (index * self.element_size)
        size: int = self.element_size
        data: bytes = val.to_bytes(self.element_size, "little", signed=getattr(self.typ, "signed", True))
        write_mem(data, self.element_size, addr)

    def __len__(self) -> int:
        return self.len

    def __repr__(self) -> str:
        return f"<array {self.typ.__name__}[{self.len}] at {hex(self.address)} of size {self.total_size} bytes>"

    @classmethod
    def __class_getitem__(cls, *args):
        return array

    def __del__(self) -> None:
        if not self.deleted:
            global _heap_registry
            try:
                del _heap_registry[self.address]
            except KeyError:
                pass
            self.deleted = True

Array = array

class struct():
    """A C-Structure, uses format like -
    ```python
    format: dict = {
        "field1": {
            "type": long, # Type of field
            "value": None # Default value
        },
        "field2": {
            "type": pointer,
            "value": None,
            "ptr_type": short, # Pointer-Specific, Type of pointer
            "ptr_val": None # Pointer-Specific, Default Pointer type
        },
        "field3": {
            "type": array,
            "value": None,
            "array_type": char, # Array-Specific, Type of Array
            "array_len": 4 # Array-Specific, Length of Array
        }
    }
    ```
    """

    def __init__(self, structure: dict, address: Optional[int]=None) -> None:
        self.struct = structure
        self.offsets = {}
        self.size = 0

        for name, info in self.struct.items():
            typ = info["type"]
            align_to = getattr(typ, "size", 1)
            self.size += align(self.size, align_to)
            self.offsets[name] = self.size
            
            field_size = 0

            # Handle
            # Arrays
            if typ == array:
                arr_typ = info.get("array_type", char)
                arr_len = info.get("array_len", 1)
                field_size = arr_typ.size * arr_len
            else:
                field_size = getattr(typ, "size", 1)
            self.size += field_size

        self.address = 0
        self.deleted = True
        if not address:
            # Allocate
            global _heap_nextalloc
            self.address = _heap_nextalloc
            _heap_nextalloc += self.size
            _heap_registry[self.address] = self.size
            write_mem(b"\x00" * self.size, self.size, self.address)
            self.deleted = False
        else:
            self.address = address

        # write defaults only if newly allocated
        if not address:
            for name, info in self.struct.items():
                if "value" in info and info["value"] is not None:
                    self.__setattr__(name, info["value"])

    def __getattr__(self, name: str) -> Any:
        global _heap_nextalloc
        if name not in self.struct: raise ValueError(f"Unknown field [{name}]")

        info = self.struct[name]
        offset = self.offsets[name]
        typ = info["type"]

        if typ == array:
            arr_typ = info.get("array_type", char)
            arr_len = info.get("array_len", 1)
            arr_elementsize = getattr(arr_typ, "size", 1)
            arr = typ(arr_typ, arr_len)
            _heap_nextalloc -= arr_len * arr_elementsize # Remove it's allocation
            arr.address = self.address + offset
            return arr
        elif typ == pointer:
            ptr_typ = info.get("ptr_type", void)
            raw = read_mem(pointer.size, self.address + offset)
            val = int.from_bytes(raw, "little", signed=getattr(ptr_typ, "signed", False))
            return typ(val)
        else:
            raw = read_mem(typ.size, self.address + offset)
            val = int.from_bytes(raw, "little", signed=getattr(typ, "signed", False))
            return typ(val)

    def __setattr__(self, name: str, value: Union[int, CBaseType, array, pointer, bytes, str]) -> None:
        if name in ("struct", "offsets", "size", "address", "deleted"):
            super().__setattr__(name, value)
            return

        if name not in self.struct:
            raise AttributeError(f"No such field '{name}'")

        info = self.struct[name]
        offset = self.offsets[name]
        typ = info["type"]

        if typ == array:
            mk_arr = False
            val = b""
            if isinstance(value, bytes):
                val = value
                mk_arr = True
            elif isinstance(value, str):
                val = value.encode("utf-8")
                mk_arr = True

            if mk_arr:
                arr = array(info.get("array_type", char), info.get("array_len", 1), self.address + offset) # Creates the arr
                for i in range(arr.len):
                    b = b""
                    if i >= len(val):
                        b = 0
                    else:
                        b = val[i]
                    arr[i] = b
                off = 0
                for i in range(arr.len):
                    arrval = arr[i]
                    if isinstance(arrval, char):
                        arrval = chr(arrval.value).encode("utf-8")
                    else:
                        arrval = arrval.value
                    write_mem(arrval, arr.element_size, self.address + offset + off)
                    off += arr.element_size
                return
            raise TypeError("Cannot directly assign to array field; use indexing.")
        elif typ == pointer:
            if isinstance(value, int):
                value = pointer(value, info.get("ptr_type", void)) # Fix
            val = value.ptr_addr
            write_mem(val.to_bytes(8, "little"), 8, self.address + offset)
        else:
            val = getattr(value, "value", value)
            write_mem(val.to_bytes(typ.size, "little", signed=typ.signed), typ.size, self.address + offset)

    def __repr__(self) -> str:
        field_str = []
        for name, info in self.struct.items():
            val = getattr(self, name)
            if isinstance(val, (array, pointer)):
                field_str.append(f"{name}={val}")
            else:
                field_str.append(f"{name}={val.value}")
        return f"<struct {', '.join(field_str)} at {hex(self.address)}>"

    @classmethod
    def __class_getitem__(cls, *args):
        return struct

    @classmethod
    def from_address(cls, structure: dict, address: int):
        """Cast an existing memory address to a struct."""
        return cls(structure, address)

    def __del__(self) -> None:
        if not self.deleted:
            global _heap_registry
            del _heap_registry[self.address]
            self.deleted = True

Struct = struct

class enum:
    """C-style enum"""
    def __init__(self, name: str, fields: dict[str, int]):
        self.name = name
        self.fields = fields
        # Auto-fill missing values like C does
        value_counter = 0
        for k, v in list(fields.items()):
            if v is None:
                fields[k] = value_counter
            value_counter = fields[k] + 1

    def __getattr__(self, key):
        if key in self.fields:
            return self.fields[key]
        raise AttributeError(f"Enum {self.name} has no field {key}")

    def __repr__(self):
        items = ", ".join([f"{k}={v}" for k, v in self.fields.items()])
        return f"<enum {self.name}: {items}>"

    @classmethod
    def __class_getitem__(cls, *args):
        return enum

# Helpers
def cstring(py_str: str) -> pointer:
    """Allocates a null-terminated C string on the heap and returns a pointer to it."""
    # Encode to bytes + null-terminator
    encoded = py_str.encode("utf-8") + b"\x00"
    size = len(encoded)

    # Allocate memory using heap-allocater
    global _heap_nextalloc, _heap_registry
    addr = _heap_nextalloc
    _heap_nextalloc += size
    _heap_registry[addr] = size

    write_mem(encoded, size, addr)
    return pointer(addr, char)
